from typing import Tuple, List, Union, Callable, Optional

import numpy as np

from ._birth_death_chain import BirthDeathChain
from ._double_well import DoubleWellDiscrete
from ._ellipsoids import Ellipsoids
from ._pbf_simulator import PBFSimulator
from ._drunkards_walk_simulator import DrunkardsWalk
from ._bickley_simulator import BickleyJet, BickleyJetDataset
from ._systems import TimeDependentSystem, TimeIndependentSystem, CustomSystem


def double_well_discrete():
    r"""MCMC process in a symmetric double well potential, spatially discretized to 100 bins.
    The discrete trajectory contains 100000 steps, discrete time step dt=10. The result object allows access to
    discretizations of varying quality as well as gives opportunity to synthetically generate more data.

    .. plot:: datasets/plot_double_well_discrete.py

    Returns
    -------
    dataset : deeptime.data.DoubleWellDiscrete
        an object that contains a markov state model corresponding to the process and the discrete trajectory
    """
    return DoubleWellDiscrete()


def ellipsoids(laziness: float = 0.97, seed=None):
    r""" Example data of a two-state markov chain which can be featurized into two parallel ellipsoids and optionally
    rotated into higher-dimensional space.

    In particular, a synthetic trajectory of observations of states :math:`S = \{0, 1\}` can be generated from
    a :class:`MSM <deeptime.markov.msm.MarkovStateModel>`. The transition probabilities have to be chosen so that
    the chain is lazy, i.e., it is more likely to stay in one state than to transition to another.

    Optionally, a continuous observation chain can be generated with two parallel ellipsoidal multivariate normal
    distributions. In this case, the MSM acts as hidden markov state model with a Gaussian output model. For
    benchmark and demonstration purposes, this observation chain can be rotated into a higher dimensional space
    and equipped with additional noise.

    .. plot:: datasets/plot_ellipsoids.py

    Parameters
    ----------
    laziness : float in half-open interval (0.5, 1.], default=0.97
        The probability to stay in either state rather than transitioning. This yields a transition matrix of

        .. math:: P = \begin{pmatrix} \lambda & 1-\lambda \\ 1-\lambda & \lambda \end{pmatrix},

        where :math:`\lambda` is the selected laziness parameter.
    seed : int, optional, default=None
        Optional random seed for reproducibility.

    Returns
    -------
    dataset : deeptime.data.Ellipsoids
        an object that contains methods to create discrete and continuous observations

    Examples
    --------
    >>> import deeptime as dt
    >>> feature_trajectory = dt.data.ellipsoids(seed=17).observations(n_steps=500)
    >>> assert feature_trajectory.shape == (500, 2)
    """
    return Ellipsoids(laziness=laziness, seed=seed)


def position_based_fluids(n_burn_in=5000, initial_positions=None, n_jobs=None):
    r""" Creates a position based fluids (PBF) simulator. It was introduced in :footcite:`macklin2013position`.
    Up to numerics the simulation is deterministic.

    The simulation box has dimensions :math:`[-40, 40]\times [-25, 25]` and the initial positions of the particles are
    around the top boundary of the box. For simplicity of use, the initial positions are fixed in this method and yield
    972 particles. For custom positioning, please use the simulator directly.

    The interaction distance is set to :math:`d = 1.5` and `n_burn_in` steps are
    performed to equilibrate the system before returning the simulator.

    For more details see :class:`PBFSimulator <deeptime.data.PBFSimulator>`.

    .. plot::

        import matplotlib.pyplot as plt
        import deeptime

        ftraj = deeptime.data.position_based_fluids(n_burn_in=150).run(300)
        f, axes = plt.subplots(3, 2, figsize=(15, 10))
        for i, ax in enumerate(axes.flat):
            ax.scatter(*(ftraj[i*50].reshape(-1, 2).T))
            ax.set_title("t = {}".format(i*50))
            ax.grid()
        f.suptitle(r'PBF dataset observations.')
        plt.show()

    Parameters
    ----------
    n_burn_in : int, default=5000
        Number of steps without any drift force to equilibrate the system.
    initial_positions : ndarray, optional, default=None
        Explicit initial positions, optional.
    n_jobs : int or None, default=None
        Number of threads to use for simulation.

    Returns
    -------
    simulator : deeptime.data.PBFSimulator
        The PBF simulator.

    References
    ----------
    .. footbibliography::
    """
    from deeptime.util.parallel import handle_n_jobs
    n_jobs = handle_n_jobs(n_jobs)
    interaction_distance = 1.5
    if initial_positions is None:
        init_pos_x = np.arange(-24, 24, interaction_distance * .9).astype(np.float32)
        init_pos_y = np.arange(-12, 24, interaction_distance * .9).astype(np.float32)
        initial_positions = np.dstack(np.meshgrid(init_pos_x, init_pos_y)).reshape(-1, 2)
    domain = np.array([80, 50])
    pbf = PBFSimulator(domain_size=domain, initial_positions=initial_positions,
                       interaction_distance=interaction_distance,
                       n_jobs=n_jobs)
    # equilibrate
    pbf.run(n_burn_in, 0)

    return pbf


def drunkards_walk(grid_size: Tuple[int, int] = (10, 10),
                   bar_location: Union[Tuple[int, int], List[Tuple[int, int]]] = (9, 9),
                   home_location: Union[Tuple[int, int], List[Tuple[int, int]]] = (0, 0)):
    r"""This example dataset simulates the steps a drunkard living in a two-dimensional plane takes finding
    either the bar or the home as two absorbing states.

    The drunkard can take steps in a 3x3 grid with uniform probability (as possible, in the corners the only
    possibilities are the ones that do not lead out of the grid). The transition matrix
    :math:`P\in\mathbb{R}^{nm\times nm}`  possesses one absorbing state for home and bar, respectively,
    and uniform two-dimensional jump probabilities in between. The grid is of size :math:`n\times m` and a point
    :math:`(i,j)` is identified with state :math:`i+nj` in the transition matrix.

    .. plot:: datasets/plot_drunkards_walk.py

    Parameters
    ----------
    grid_size : tuple
        The grid size, must be tuple of length two.
    bar_location : tuple
        The bar location, must be valid coordinate and tuple of length two.
    home_location : tuple
        The home location, must be valid coordinate and tuple of length two.

    Returns
    -------
    simulator : deeptime.data.DrunkardsWalk
        Simulator instance.
    """
    return DrunkardsWalk(grid_size, bar_location=bar_location, home_location=home_location)


def bickley_jet(n_particles: int, n_jobs: Optional[int] = None, seed: Optional[int] = None) -> BickleyJetDataset:
    r"""Simulates the Bickley jet for a number of particles.
    The implementation is based on :footcite:`hadjighasem2016spectral` with parameters

    .. math::

            \begin{aligned}
                U_0 &= 5.4138 \times \frac{10^6\mathrm{m}}{\mathrm{day}},\\
                L_0 &= 1.77 \times 10^6\,\mathrm{m},\\
                r_0 &= 6.371 \times 10^6\,\mathrm{m},\\
                c &= (0.1446, 0.205, 0.461)^\top U_0,\\
                \mathrm{eps} &= (0.075, 0.15, 0.3)^\top,\\
                k &= (2,4,6)^\top \frac{1}{r_0},
            \end{aligned}

    in a domain :math:`\Omega = [0, 20] \times [-3, 3]`. The resulting dataset describes the temporal evolution
    of :code:`n_particles` over 401 timesteps in :math:`\Omega`. The domain is periodic in x-direction.
    The dataset offers methods to wrap the domain into three-dimensional
    space onto the surface of a cylinder

    .. math::

        \begin{pmatrix} x \\ y \end{pmatrix} \mapsto \begin{pmatrix}
            r\cdot \cos\left( 2\pi \frac{x}{20} \right) \\
            r\cdot \sin\left( 2\pi \frac{x}{20} \right) \\
            \frac{y}{3}
        \end{pmatrix},

    with the option to further discretize the three-dimensional dataspace via binning. This way the
    discontinuity introduced by 2D periodicity is treated.

    .. plot::

        import matplotlib.pyplot as plt
        import deeptime

        n_particles = 1000
        dataset = deeptime.data.bickley_jet(n_particles, n_jobs=8)

        fig, axes = plt.subplots(2, 3, sharex=True, sharey=True, figsize=(16, 10))

        for t, ax in zip([0, 1, 2, 200, 300, 400], axes.flatten()):
            ax.scatter(*dataset[t].T, c=dataset[0, :, 0], s=50)
            ax.set_title(f"Particles at t={t}")

    Parameters
    ----------
    n_particles : int
        Number of particles which are propagated.
    n_jobs : int or None, default=None
        Number of threads to use for simulation.
    seed : int or None, optional, default=None
        Random seed used for initialization of particle positions at :math:`t=0`.

    Returns
    -------
    dataset : BickleyJetDataset
        Dataset over all the generated frames.

    See Also
    --------
    BickleyJet
        Underlying trajectory generator.

    Examples
    --------

    >>> import deeptime
    >>> dataset = deeptime.data.bickley_jet(n_particles=5, n_jobs=1)
    >>> # shape is 401 frames for 5 particles in two dimensions
    >>> print(dataset.data.shape)
    (401, 5, 2)

    >>> # returns a timelagged dataset for first and last frame
    >>> endpoints = dataset.endpoints_dataset()
    >>> endpoints.data.shape
    (5, 2)

    >>> # maps the endpoints dataset onto a cylinder of radius 5
    >>> endpoints_3d = endpoints.to_3d(radius=5.)
    >>> endpoints_3d.data.shape
    (5, 3)

    >>> # bins the data uniformly with 10 bins per axis
    >>> endpoints_3d_clustered = endpoints_3d.cluster(n_bins=10)
    >>> # 5 particles and 10*10*10 bins
    >>> endpoints_3d_clustered.data.shape
    (5, 1000)

    References
    ----------
    .. footbibliography::
    """
    from deeptime.util.parallel import handle_n_jobs
    n_jobs = handle_n_jobs(n_jobs)
    simulator = BickleyJet(h=1e-2, n_steps=10)
    traj = simulator.generate(n_particles=n_particles, n_jobs=n_jobs, seed=seed)
    traj_reshaped = traj.transpose(1, 0, 2)
    return BickleyJetDataset(traj_reshaped)


def birth_death_chain(q, p, sparse=False):
    r""" Generates a birth and death chain simulator from annihilation and creation probabilities `q` and `p`.

    A general birth and death chain on a d-dimensional state space has the transition matrix

    .. math::

        p_{ij} = \begin{cases}
            q_i &\text{, if } j=i-1 \text{ and } i>0,\\
            r_i &\text{, if } j=i,\\
            p_i &\text{, if } j=i+1 \text{ and } i < d-1
        \end{cases}


    The annihilation probability of state :math:`i=1` must not be zero, same for the creation probability
    of the last state :math:`i=n`. The sum of the probabilities must be bounded component-wise, i.e.,
    :math:`q_i + p_i \leq 1\;\forall i=1,\ldots ,n`.

    .. plot:: datasets/plot_birth_death_chain.py

    Parameters
    ----------
    q : array_like
        Annihilation probabilities for transition from i to i-1.
    p : array_like
        Creation probabilities for transition from i to i+1.
    sparse : bool, optional, default=False
        Whether to use sparse matrices.

    Returns
    -------
    chain : deeptime.data.BirthDeathChain
        The chain.
    """
    return BirthDeathChain(q, p, sparse=sparse)


def tmatrix_metropolis1d(energies, d=1.0):
    r"""Transition matrix describing the Metropolis chain jumping
    between neighbors in a discrete 1D energy landscape.

    .. plot:: datasets/plot_tmatrix_1d.py

    Parameters
    ----------
    energies : (M,) ndarray
        Energies in units of kT
    d : float (optional)
        Diffusivity of the chain, d in (0, 1]

    Returns
    -------
    P : (M, M) ndarray
        Transition matrix of the Markov chain

    Notes
    -----
    Transition probabilities are computed as

    .. math::

        \begin{aligned}
        p_{i,i-1} &= 0.5 d \min \left\{ 1.0, \mathrm{e}^{-(E_{i-1} - E_i)} \right\}, \\
        p_{i,i+1} &= 0.5 d \min \left\{ 1.0, \mathrm{e}^{-(E_{i+1} - E_i)} \right\}, \\
        p_{i,i}   &= 1.0 - p_{i,i-1} - p_{i,i+1}.
        \end{aligned}

    """
    import math
    # check input
    assert 0 < d <= 1, f'Diffusivity must be in (0,1]. Trying to set the invalid value {d}.'
    # init
    n = len(energies)
    transition_matrix = np.zeros((n, n))
    # set off diagonals
    transition_matrix[0, 1] = 0.5 * d * min(1.0, math.exp(-(energies[1] - energies[0])))
    for i in range(1, n - 1):
        transition_matrix[i, i - 1] = 0.5 * d * min(1.0, math.exp(-(energies[i - 1] - energies[i])))
        transition_matrix[i, i + 1] = 0.5 * d * min(1.0, math.exp(-(energies[i + 1] - energies[i])))
    transition_matrix[n - 1, n - 2] = 0.5 * d * min(1.0, math.exp(-(energies[n - 2] - energies[n - 1])))
    # normalize
    transition_matrix += np.diag(1.0 - np.sum(transition_matrix, axis=1))
    # done
    return transition_matrix


def sqrt_model(n_samples, seed=None):
    r""" Sample a hidden state and an sqrt-transformed emission trajectory.
    We sample a hidden state trajectory and sqrt-masked emissions in two
    dimensions such that the two metastable states are not linearly separable.

    .. plot:: datasets/plot_sqrt_model.py

    Parameters
    ----------
    n_samples : int
        Number of samples to produce.
    seed : int, optional, default=None
        Random seed to use. Defaults to None, which means that the random device will be default-initialized.

    Returns
    -------
    sequence : (n_samples, ) ndarray
        The discrete states.
    trajectory : (n_samples, ) ndarray
        The observable.

    Notes
    -----
    First, the hidden discrete-state trajectory is simulated. Its transition matrix is given by

    .. math::
        P = \begin{pmatrix}0.95 & 0.05 \\ 0.05 & 0.95 \end{pmatrix}.

    The observations are generated via the means are :math:`\mu_0 = (0, 1)^\top` and :math:`\mu_1= (0, -1)^\top`,
    respectively, as well as the covariance matrix

    .. math::
        C = \begin{pmatrix} 30 & 0 \\ 0 & 0.015 \end{pmatrix}.

    Afterwards, the trajectory is transformed via

    .. math::
        (x, y) \mapsto (x, y + \sqrt{| x |}).
    """
    from deeptime.markov.msm import MarkovStateModel

    state = np.random.RandomState(seed)
    cov = sqrt_model.cov
    states = sqrt_model.states
    msm = MarkovStateModel(sqrt_model.transition_matrix)
    dtraj = msm.simulate(n_samples, seed=seed)
    traj = states[dtraj, :] + state.multivariate_normal(np.zeros(len(cov)), cov, size=len(dtraj),
                                                        check_valid='ignore')
    traj[:, 1] += np.sqrt(np.abs(traj[:, 0]))
    return dtraj, traj


sqrt_model.cov = np.array([[30.0, 0.0], [0.0, 0.015]])
sqrt_model.states = np.array([[0.0, 1.0], [0.0, -1.0]])
sqrt_model.transition_matrix = np.array([[0.95, 0.05], [0.05, 0.95]])


def swissroll_model(n_samples, seed=None):
    r""" Sample a hidden state and an swissroll-transformed emission trajectory, so that the states are not
    linearly separable.

    .. plot::

        import matplotlib.pyplot as plt
        from matplotlib import animation
        from deeptime.data import swissroll_model

        n_samples = 15000
        dtraj, traj = swissroll_model(n_samples)

        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')
        ax.scatter(*traj.T, marker='o', s=20, c=dtraj, alpha=0.6)

    Parameters
    ----------
    n_samples : int
        Number of samples to produce.
    seed : int, optional, default=None
        Random seed to use. Defaults to None, which means that the random device will be default-initialized.

    Returns
    -------
    sequence : (n_samples, ) ndarray
        The discrete states.
    trajectory : (n_samples, ) ndarray
        The observable.

    Notes
    -----
    First, the hidden discrete-state trajectory is simulated. Its transition matrix is given by

    .. math::
        P = \begin{pmatrix}0.95 & 0.05 &      &      \\ 0.05 & 0.90 & 0.05 &      \\
                                & 0.05 & 0.90 & 0.05 \\      &      & 0.05 & 0.95 \end{pmatrix}.

    The observations are generated via the means are :math:`\mu_0 = (7.5, 7.5)^\top`, :math:`\mu_1= (7.5, 15)^\top`,
    :math:`\mu_2 = (15, 15)^\top`, and :math:`\mu_3 = (15, 7.5)^\top`,
    respectively, as well as the covariance matrix

    .. math::
        C = \begin{pmatrix} 1 & 0 \\ 0 & 1 \end{pmatrix}.

    Afterwards, the trajectory is transformed via

    .. math::
        (x, y) \mapsto (x \cos (x), y, x \sin (x))^\top.
    """
    from deeptime.markov.msm import MarkovStateModel

    state = np.random.RandomState(seed)
    cov = swissroll_model.cov
    states = swissroll_model.states
    msm = MarkovStateModel(swissroll_model.transition_matrix)
    dtraj = msm.simulate(n_samples, seed=seed)
    traj = states[dtraj, :] + state.multivariate_normal(np.zeros(len(cov)), cov, size=len(dtraj),
                                                        check_valid='ignore')
    x = traj[:, 0]
    return dtraj, np.vstack([x * np.cos(x), traj[:, 1], x * np.sin(x)]).T


swissroll_model.cov = np.array([[1.0, 0.0], [0.0, 1.0]])
swissroll_model.states = np.array([[7.5, 7.5], [7.5, 15.0], [15.0, 15.0], [15.0, 7.5]])
swissroll_model.transition_matrix = np.array([[0.95, 0.05, 0.00, 0.00], [0.05, 0.90, 0.05, 0.00],
                                              [0.00, 0.05, 0.90, 0.05], [0.00, 0.00, 0.05, 0.95]])


def quadruple_well(h: float = 1e-3, n_steps: int = 10000):
    r""" This dataset generates trajectories of a two-dimensional particle living in a quadruple-well potential
    landscape. It is subject to the stochastic differential equation

    .. math::

        \mathrm{d}X_t = \nabla V(X_t) \mathrm{d}t + \sqrt{2\beta^{-1}}\mathrm{d}W_t

    with :math:`W_t` being a Wiener process and the potential :math:`V` being given by

    .. math::

        V(x) = (x_1 - 1)^2 + (x_2 - 1)^2.

    The inverse temperature is set to be :math:`\beta = 4`.

    .. plot:: datasets/plot_quadruple_well.py

    Parameters
    ----------
    h : float, default = 1e-3
        Integration step size. The implementation uses an Euler-Maruyama integrator.
    n_steps : int, default = 10000
        Number of integration steps between each evaluation. That means the default lag time is :code:`h*n_steps=10`.

    Returns
    -------
    model : TimeIndependentSystem
        The model.

    Examples
    --------
    The model possesses the capability to simulate trajectories as well as be evaluated at test points:

    >>> import numpy as np
    >>> import deeptime as dt

    First, set up the model (which internally already creates the integrator).

    >>> model = dt.data.quadruple_well(h=1e-3, n_steps=100)  # create model instance

    Now, a trajectory can be generated:

    >>> traj = model.trajectory([0., 0.], 1000, seed=42, n_jobs=1)  # simulate trajectory
    >>> assert traj.shape == (1000, 2)  # 1000 evaluations from initial condition [0, 0]

    Or, alternatively the model can be evaluated at test points (mapping forward using the dynamical system):

    >>> test_points = np.random.uniform(-2, 2, (100, 2))  # 100 test point in [-2, 2] x [-2, 2]
    >>> evaluations = model(test_points, seed=53, n_jobs=1)
    >>> assert evaluations.shape == (100, 2)
    """
    from ._data_bindings import QuadrupleWell2D
    return TimeIndependentSystem(QuadrupleWell2D(), h, n_steps)


def quadruple_well_asymmetric(h=1e-3, n_steps=10000):
    r""" This dataset generates trajectories of a two-dimensional particle living in an asymmetric quadruple-well
    potential landscape. It is subject to the stochastic differential equation

    .. math::

        \mathrm{d}X_t = \nabla V(X_t) \mathrm{d}t + \sigma\mathrm{d}W_t

    with :math:`W_t` being a Wiener process and the potential :math:`V` being given by

    .. math::

        V(x) = (x_1^4-\frac{x_1^3}{16}-2x_1^2+\frac{3x_1}{16}) + (x_2^4-\frac{x_1^3}{8}-2x_1^2+\frac{3x_1}{8}).

    The stochastic force parameter is set to :math:`\sigma = 0.6`.

    .. plot:: datasets/plot_quadruple_well_asymmetric.py

    Parameters
    ----------
    h : float, default = 1e-3
        Integration step size. The implementation uses an Euler-Maruyama integrator.
    n_steps : int, default = 10000
        Number of integration steps between each evaluation. That means the default lag time is :code:`h*n_steps=10`.

    Returns
    -------
    model : TimeIndependentSystem
        The model.

    Examples
    --------
    The model possesses the capability to simulate trajectories as well as be evaluated at test points:

    >>> import numpy as np
    >>> import deeptime as dt

    First, set up the model (which internally already creates the integrator).

    >>> model = dt.data.quadruple_well_asymmetric(h=1e-3, n_steps=100)  # create model instance

    Now, a trajectory can be generated:

    >>> traj = model.trajectory(np.array([[0., 0.]]), 1000, seed=42, n_jobs=1)  # simulate trajectory
    >>> assert traj.shape == (1000, 2)  # 1000 evaluations from initial condition [0, 0]

    Or, alternatively the model can be evaluated at test points (mapping forward using the dynamical system):

    >>> test_points = np.random.uniform(-2, 2, (100, 2))  # 100 test point in [-2, 2] x [-2, 2]
    >>> evaluations = model(test_points, seed=53, n_jobs=1)
    >>> assert evaluations.shape == (100, 2)
    """
    from ._data_bindings import QuadrupleWellAsymmetric2D
    return TimeIndependentSystem(QuadrupleWellAsymmetric2D(), h, n_steps)


def triple_well_2d(h=1e-5, n_steps=10000):
    r""" This dataset generates trajectories of a two-dimensional particle living in a triple-well potential
    landscape. The example can be found in :footcite:`schutte2013metastability`.

    The particle is subject to the stochastic differential equation

    .. math::

        \mathrm{d}X_t = \nabla V(X_t) \mathrm{d}t + \sigma(t, X_t)\mathrm{d}W_t

    with :math:`W_t` being a Wiener process, :math:`\sigma = 1.09`, and the potential :math:`V` being given by

    .. math::

        \begin{aligned}
        V(x) &= 3e^{-x^2 - (y-\frac{1}{3})^2} - 3e^{-x^2 - (y - \frac{5}{3})^2} \\
        &\quad - 5e^{-(x-1)^2 - y^2} - 5e^{-(x+1)^2 - y^2} \\
        &\quad + \frac{2}{10} x^4 + \frac{2}{10}\left(y-\frac{1}{3}\right)^4.
        \end{aligned}

    .. plot:: datasets/plot_triple_well_2d.py

    Parameters
    ----------
    h : float, default = 1e-5
        Integration step size. The implementation uses an Euler-Maruyama integrator.
    n_steps : int, default = 10000
        Number of integration steps between each evaluation. That means the default lag time is :code:`h*n_steps=0.1`.

    Returns
    -------
    model : TimeIndependentSystem
        The model.

    Examples
    --------
    The model possesses the capability to simulate trajectories as well as be evaluated at test points:

    >>> import numpy as np
    >>> import deeptime as dt

    First, set up the model (which internally already creates the integrator).

    >>> model = dt.data.triple_well_2d(h=1e-3, n_steps=100)  # create model instance

    Now, a trajectory can be generated:

    >>> traj = model.trajectory(np.array([[-1., 0.]]), 1000, seed=42, n_jobs=1)  # simulate trajectory
    >>> assert traj.shape == (1000, 2)  # 1000 evaluations from initial condition [0, 0]

    Or, alternatively the model can be evaluated at test points (mapping forward using the dynamical system):

    >>> test_points = np.random.uniform(-2, 2, (100, 2))  # 100 test point in [-2, 2] x [-2, 2]
    >>> evaluations = model(test_points, seed=53, n_jobs=1)
    >>> assert evaluations.shape == (100, 2)

    References
    ----------
    .. footbibliography::
    """
    from ._data_bindings import TripleWell2D
    return TimeIndependentSystem(TripleWell2D(), h, n_steps)


def abc_flow(h=1e-3, n_steps=10000):
    r""" The Arnold-Beltrami-Childress flow. :footcite:`arnold1966topology`
    It is generated by the ODE

    .. math::

        \begin{aligned}
        \dot{x} &= A\sin(z) + C\cos(y)\\
        \dot{y} &= B\sin(x) + A\cos(z)\\
        \dot{z} &= C\sin(y) + B\cos(x)
        \end{aligned}

    on the domain :math:`\Omega=[0, 2\pi]^3` with the parameters :math:`A=\sqrt{3}`, :math:`B=\sqrt{2}`,
    and :math:`C=1`.

    .. plot::

        import numpy as np
        import deeptime as dt
        import matplotlib.pyplot as plt
        from mpl_toolkits.mplot3d import Axes3D

        system = dt.data.abc_flow(n_steps=100)
        scatters = [np.random.uniform(np.pi-.5, np.pi+.5, size=(200, 3))]
        for _ in range(12):
            scatters.append(system(scatters[-1], n_jobs=8))

        f = plt.figure(figsize=(18, 18))
        f.suptitle('Evolution of test points in the ABC flowfield')
        for i in range(4):
            for j in range(3):
                ix = j + i*3
                ax = f.add_subplot(4, 3, ix+1, projection='3d')
                ax.set_title(f"T={ix*system.n_steps*system.h:.2f}")
                ax.scatter(*scatters[ix].T, c=np.linspace(0, 1, num=200))
                ax.set_xlim([0, 2*np.pi])
                ax.set_ylim([0, 2*np.pi])
                ax.set_zlim([0, 2*np.pi])
        plt.show()


    Parameters
    ----------
    h : float, default = 1e-3
        Integration step size. The implementation uses an Runge-Kutta integrator.
    n_steps : int, default = 10000
        Number of integration steps between each evaluation. That means the default lag time is :code:`h*n_steps=10`.

    Returns
    -------
    system : TimeIndependentSystem
        The system.

    Examples
    --------
    The model possesses the capability to simulate trajectories as well as be evaluated at test points:

    >>> import numpy as np
    >>> import deeptime as dt

    First, set up the model (which internally already creates the integrator).

    >>> model = dt.data.abc_flow(h=1e-3, n_steps=100)  # create model instance

    Now, a trajectory can be generated:

    >>> traj = model.trajectory(np.array([[-1., 0., 0.]]), 1000, seed=42, n_jobs=1)  # simulate trajectory
    >>> assert traj.shape == (1000, 3)  # 1000 evaluations from initial condition [0, 0]

    Or, alternatively the model can be evaluated at test points (mapping forward using the dynamical system):

    >>> test_points = np.random.uniform(.5, 1.5, (100, 3))  # 100 test points
    >>> evaluations = model(test_points, seed=53, n_jobs=1)
    >>> assert evaluations.shape == (100, 3)

    References
    ----------
    .. footbibliography::
    """
    from ._data_bindings import ABCFlow
    return TimeIndependentSystem(ABCFlow(), h, n_steps)


def ornstein_uhlenbeck(h=1e-3, n_steps=500):
    r""" The one-dimensional Ornstein-Uhlenbeck process. It is given by the stochastic differential equation

    .. math::

        dX_t = -\alpha X_t dt + \sqrt{2\beta^{-1}}dW_t

    with parameters :math:`\alpha=1` and :math:`\beta=4`.

    .. plot:: datasets/plot_ornstein_uhlenbeck.py

    Parameters
    ----------
    h : float, default = 1e-3
        Integration step size. The implementation uses an Euler-Maruyama integrator.
    n_steps : int, default = 500
        Number of integration steps between each evaluation. That means the default lag time is :code:`h*n_steps=10`.

    Returns
    -------
    system : TimeIndependentSystem
        The system.

    Examples
    --------
    The model possesses the capability to simulate trajectories as well as be evaluated at test points:

    >>> import numpy as np
    >>> import deeptime as dt

    First, set up the model (which internally already creates the integrator).

    >>> model = dt.data.ornstein_uhlenbeck(h=1e-3, n_steps=100)  # create model instance

    Now, a trajectory can be generated:

    >>> traj = model.trajectory(np.array([[-1.]]), 1000, seed=42, n_jobs=1)  # simulate trajectory
    >>> assert traj.shape == (1000, 1)  # 1000 evaluations from initial condition [0, 0]

    Or, alternatively the model can be evaluated at test points (mapping forward using the dynamical system):

    >>> test_points = np.random.uniform(.5, 1.5, (100, 1))  # 100 test points
    >>> evaluations = model(test_points, seed=53, n_jobs=1)
    >>> assert evaluations.shape == (100, 1)
    """
    from ._data_bindings import OrnsteinUhlenbeck
    return TimeIndependentSystem(OrnsteinUhlenbeck(), h, n_steps)


def prinz_potential(h=1e-5, n_steps=500, temperature_factor=1., mass=1., damping=1.):
    r""" Particle diffusing in a one-dimensional quadruple well potential landscape. :footcite:`prinz2011markov`

    The potential is defined as

    .. math::
        V(x) = 4 \left( x^8 + 0.8 e^{-80 x^2} + 0.2 e^{-80 (x-0.5)^2} + 0.5 e^{-40 (x+0.5)^2}\right).

    The integrator is an Euler-Maruyama type integrator, updating the current state :math:`x_t` via

    .. math::
        x_{t+1} =  x_t - \frac{h\nabla V(x_t)}{m\cdot d} + \sqrt{2\frac{h\cdot\mathrm{kT}}{m\cdot d}}\eta_t,

    where :math:`m` is the mass, :math:`d` the damping factor, and :math:`\eta_t \sim \mathcal{N}(0, 1)`.

    The locations of the minima can be accessed via the `minima` attribute.

    .. plot:: datasets/plot_prinz.py

    Parameters
    ----------
    h : float, default=1e-5
        The integrator step size. If the temperature is too high and the step size too large, the
        integrator may produce NaNs.
    n_steps : int, default=500
        Number of integration steps between each evaluation of the system's state.
    temperature_factor : float, default=1
        The temperature kT.
    mass : float, default=1
        The particle's mass.
    damping : float, default=1
        Damping factor.

    Returns
    -------
    system : TimeIndependentSystem
        The system.

    Examples
    --------
    The model possesses the capability to simulate trajectories as well as be evaluated at test points:

    >>> import numpy as np
    >>> from deeptime.data import prinz_potential

    First, set up the model (which internally already creates the integrator).

    >>> model = prinz_potential(h=1e-3, n_steps=100)  # create model instance

    Now, a trajectory can be generated:

    >>> traj = model.trajectory(np.array([[-0.]]), 1000, seed=42, n_jobs=1)  # simulate trajectory
    >>> assert traj.shape == (1000, 1)  # 1000 evaluations from initial condition [0]

    Or, alternatively the model can be evaluated at test points (mapping forward using the dynamical system):

    >>> test_points = np.random.uniform(.5, 1.5, (100, 1))  # 100 test points
    >>> evaluations = model(test_points, seed=53, n_jobs=1)
    >>> assert evaluations.shape == (100, 1)

    References
    ----------
    .. footbibliography::
    """
    from ._data_bindings import Prinz
    system = TimeIndependentSystem(Prinz(), h, n_steps, props={
        'kT': temperature_factor,
        'mass': mass,
        'damping': damping
    })
    system.minima = [-0.73943018, -0.22373758, 0.26914935, 0.67329636]
    return system


def triple_well_1d(h=1e-3, n_steps=500):
    r""" A simple one-dimensional triple-well potential landscape. It is given by the stochastic differential equation

    .. math::

        \mathrm{d}X_t = \nabla V(X_t) \mathrm{d}t + \sigma(t, X_t)\mathrm{d}W_t

    with :math:`W_t` being a Wiener process, :math:`\sigma = 1.09`, and the potential :math:`V` being given by

    .. math::

        V(x) = 5 - 24.82 x + 41.4251 x^2 - 27.5344 x^3 + 8.53128 x^4 - 1.24006 x^5 + 0.0684 x^6.

    .. plot:: datasets/plot_triple_well_1d.py

    Parameters
    ----------
    h : float, default=1e-3
        Integration step size.
    n_steps : int, default=500
        Default number of integration steps per evaluation.

    Returns
    -------
    system : TimeIndependentSystem
        The system.
    """
    from ._data_bindings import TripleWell1D
    return TimeIndependentSystem(TripleWell1D(), h, n_steps)


def double_well_2d(h=1e-3, n_steps=10000):
    r""" This dataset generates trajectories of a two-dimensional particle living in a double-well potential
    landscape.

    The particle is subject to the stochastic differential equation

    .. math::

        \mathrm{d}X_t = \nabla V(X_t) \mathrm{d}t + \sigma(t, X_t)\mathrm{d}W_t

    with :math:`W_t` being a Wiener process, :math:`\sigma = 0.7`, and the potential :math:`V` being given by

    .. math::

        V(x) = (x_1^2 - 1)^2 + x_2^2.

    .. plot:: datasets/plot_double_well_2d.py

    Parameters
    ----------
    h : float, default = 1e-5
        Integration step size. The implementation uses an Euler-Maruyama integrator.
    n_steps : int, default = 10000
        Number of integration steps between each evaluation. That means the default lag time is :code:`h*n_steps=0.1`.

    Returns
    -------
    model : TimeIndependentSystem
        The model.

    Examples
    --------
    The model possesses the capability to simulate trajectories as well as be evaluated at test points:

    >>> import numpy as np
    >>> import deeptime as dt

    First, set up the model (which internally already creates the integrator).

    >>> model = dt.data.double_well_2d(h=1e-3, n_steps=100)  # create model instance

    Now, a trajectory can be generated:

    >>> traj = model.trajectory(np.array([[-1., 0.]]), 1000, seed=42, n_jobs=1)  # simulate trajectory
    >>> assert traj.shape == (1000, 2)  # 1000 evaluations from initial condition [0, 0]

    Or, alternatively the model can be evaluated at test points (mapping forward using the dynamical system):

    >>> test_points = np.random.uniform(-2, 2, (100, 2))  # 100 test point in [-2, 2] x [-2, 2]
    >>> evaluations = model(test_points, seed=53, n_jobs=1)
    >>> assert evaluations.shape == (100, 2)
    """
    from ._data_bindings import DoubleWell2D
    return TimeIndependentSystem(DoubleWell2D(), h, n_steps)


def time_dependent_quintuple_well(h=1e-3, n_steps=10000, beta=5.):
    r"""This dataset generates trajectories of a two-dimensional particle living in a time-dependent quintuple-well
    potential landscape. The potential wells are slowly oscillating around the origin.

    The dynamics are described by the potential landscape

    .. math::
        V(t, x, y) = \cos \left(s\;\mathrm{arctan}(y, x) - \frac{\pi}{2}t \right) + 10 \left( \sqrt{x^2+y^2} - \frac{3}{2} - \frac{1}{2}\sin(2\pi t) \right)^2

    subject to the SDE

    .. math::
        dX_t = -\nabla V(X_t, t) dt + \sqrt{2\beta^{-1}}dW_t,

    where :math:`\beta` is the temperature and :math:`W_t` a Wiener process.

    Parameters
    ----------
    h : float, optional, default=1e-3
        Integration step size.
    n_steps : int, optional, default=10000
        Number of steps to evaluate between recording states.
    beta : float, default=5.
        The inverse temperature.

    Returns
    -------
    system : TimeDependentSystem
        The system.

    Examples
    --------
    The model possesses the capability to simulate trajectories as well as be evaluated at test points:

    >>> import numpy as np
    >>> import deeptime as dt

    First, set up the model (which internally already creates the integrator).

    >>> model = dt.data.time_dependent_quintuple_well(h=1e-3, n_steps=100, beta=5.)  # create model instance

    Now, a trajectory can be generated:

    >>> traj = model.trajectory(0., np.array([[-1., 0.]]), 1000, seed=42, n_jobs=1)  # simulate trajectory
    >>> assert traj.shape == (1000, 2)  # 1000 evaluations from initial condition [0, 0] with t=0

    Or, alternatively the model can be evaluated at test points (mapping forward using the dynamical system):

    >>> test_points = np.random.uniform(-2, 2, (100, 2))  # 100 test point in [-2, 2] x [-2, 2]
    >>> evaluations = model(0, test_points, seed=53, n_jobs=1)  # start at t=0
    >>> assert evaluations.shape == (100, 2)
    """
    from ._data_bindings import TimeDependent5Well2D
    return TimeDependentSystem(TimeDependent5Well2D(), h, n_steps, props={'beta': beta})


def custom_sde(dim: int, rhs: Callable, sigma: np.ndarray, h: float, n_steps: int):
    r""" This function allows the definition of custom stochastic differential equations (SDEs) of the form

    .. math::

        \mathrm{d}X_t = F(X_t) \mathrm{d}t + \sigma(t, X_t)\mathrm{d}W_t,

    where the right-hand side :math:`F` should map an :code:`dim`-dimensional array-like object to an
    :code:`dim`-dimensional array-like object. The prefactor in front of the Wiener process :math:`W_t` is assumed
    to be constant with respect to time and state, i.e.,

    .. math::

        \sigma(t, X_t) = \sigma \in\mathbb{R}^{\mathrm{dim}\times\mathrm{dim}}.

    .. plot:: datasets/plot_custom_sde.py

    Parameters
    ----------
    dim : int, positive and less or equal to 5
        The dimension of the SDE's state vector :math:`X_t`. Must be less or equal to 5.
    rhs : Callable
        The right-hand side function :math:`F(X_t)`. It must map a dim-dimensional array like object to a
        dim-dimensional array or list.
    sigma : (dim, dim) ndarray
        The sigma parameter.
    h : float
        Step size for the Euler-Maruyama integrator.
    n_steps : int
        Number of integration steps per evaluation / recording of the state.

    Returns
    -------
    system : CustomSystem
        The system.

    Examples
    --------
    First, some imports.

    >>> import deeptime as dt
    >>> import numpy as np

    Then, we can define the right-hand side. Here, we choose the force of an harmonic spherical inclusion potential.

    >>> def harmonic_sphere_force(x, radius=.5, k=1.):
    ...     dist_to_origin = np.linalg.norm(x)
    ...     dist_to_sphere = dist_to_origin - radius
    ...     if dist_to_sphere > 0:
    ...         return -k * dist_to_sphere * np.array(x) / dist_to_origin
    ...     else:
    ...         return [0., 0.]

    This we can use as right-hand side to define our SDE with :math:`\sigma = \mathrm{diag}(1, 1)`.

    >>> sde = dt.data.custom_sde(dim=2, rhs=lambda x: harmonic_sphere_force(x, radius=.5, k=1),
    ...                          sigma=np.diag([1., 1.]), h=1e-3, n_steps=1)

    Here, :code:`h` is the step-size of the (Euler-Maruyama) integrator and :code:`n_steps` refers to the number of
    integration steps for each evaluation.
    Given the SDE instance, we can generate trajectories via

    >>> trajectory = sde.trajectory(x0=[[0., 0.]], length=10, seed=55)
    >>> assert trajectory.shape == (10, 2)

    or propagate (in this case 300) sample points by :code:`n_steps`:

    >>> propagated_samples = sde(np.random.normal(scale=.1, size=(300, 2)))
    >>> assert propagated_samples.shape == (300, 2)
    """
    from . import _data_bindings as bindings
    if not (isinstance(dim, int) and 0 < dim <= 5):
        raise ValueError("Dimension must be positive and at most 5.")

    sigma = np.atleast_2d(np.array(sigma).squeeze())
    sigma_shape = sigma.shape
    if not sigma_shape == (dim, dim):
        raise ValueError("Sigma must be DIM x DIM matrix but had shape", sigma_shape)

    SDE = getattr(bindings, f'PySDE{dim}D')
    return CustomSystem(SDE(sigma, rhs), h, n_steps)


def custom_ode(dim: int, rhs: Callable, h: float, n_steps: int):
    r""" This function allows the definition of custom ordinary differential equations (ODEs) of the form

    .. math::

        \mathrm{d}X_t = F(X_t) \mathrm{d}t,

    where the right-hand side :math:`F` should map an :code:`dim`-dimensional array-like object to an
    :code:`dim`-dimensional array-like object.

    .. plot:: datasets/plot_custom_ode.py

    Parameters
    ----------
    dim : int, positive and less or equal to 5
        The dimension of the SDE's state vector :math:`X_t`. Must be less or equal to 5.
    rhs : Callable
        The right-hand side function :math:`F(X_t)`. It must map a dim-dimensional array like object to a
        dim-dimensional array or list.
    h : float
        Step size for the Runge-Kutta integrator.
    n_steps : int
        Number of integration steps per evaluation / recording of the state.

    Returns
    -------
    system : CustomSystem
        The system.

    Examples
    --------
    First, some imports.

    >>> import numpy as np
    >>> import deeptime as dt

    We can define the right-hand side to model an exponential decay

    >>> def rhs(x):
    ...     return [-.5 * x[0]]

    and obtain the system

    >>> system = dt.data.custom_ode(dim=1, rhs=rhs, h=1e-3, n_steps=20)

    where :code:`n_steps` is the number of (Runge-Kutta 45) integration steps per evaluation and :code:`h` the
    step-size. With the system, one can generate trajectories

    >>> traj = system.trajectory(x0=[[1.]], length=50, seed=45)
    >>> assert traj.shape == (50, 1)

    as well as propagate sample points by :code:`n_steps`:

    >>> propagated_samples = system(np.random.uniform(size=(100, 1)))
    >>> assert propagated_samples.shape == (100 ,1)
    """
    from . import _data_bindings as bindings
    if not (isinstance(dim, int) and 0 < dim <= 5):
        raise ValueError("Dimension must be positive and at most 5.")

    ODE = getattr(bindings, f'PyODE{dim}D')
    return CustomSystem(ODE(rhs), h, n_steps)



from typing import Union

import numpy as np
import torch

from . import GaussianKernel

TensorOrArray = Union[np.ndarray, torch.Tensor]


class TorchGaussianKernel(GaussianKernel):
    r""" Gaussian kernel which has a dependency to PyTorch and can handle evaluation of Gram matrices as well as
    (over the batch-size) vectorized evaluations of the kernel.

    See Also
    --------
    GaussianKernel
    """

    def __init__(self, sigma):
        np_sigma = sigma
        if isinstance(np_sigma, torch.Tensor):
            np_sigma = sigma.detach().cpu().numpy()
        super().__init__(np_sigma)
        self._sigma = sigma
        self.impl = 'cdist'

    @staticmethod
    def cdist(x1, x2):
        x1_norm = x1.pow(2).sum(dim=-1, keepdim=True)
        x2_norm = x2.pow(2).sum(dim=-1, keepdim=True)
        res = torch.addmm(
            x2_norm.t(),
            x1,
            x2.t(),
            alpha=-2
        ).add_(x1_norm).clamp_min_(1e-16)
        return res

    def apply_torch(self, data_1: torch.Tensor, data_2: torch.Tensor):
        distance_matrix = TorchGaussianKernel.cdist(data_1, data_2)
        return torch.exp(-distance_matrix / (2. * self.sigma ** 2))

    def _evaluate(self, x, y) -> TensorOrArray:
        if isinstance(x, torch.Tensor) and isinstance(y, torch.Tensor):
            dxy_squared = torch.pow(x - y, 2).sum(-1).clamp_min_(0.)
            return torch.exp(-dxy_squared / (2. * self.sigma ** 2))
        else:
            out = []
            for batch_ix in range(len(x)):
                out.append(super()._evaluate(x[batch_ix], y[batch_ix]))
            return np.array(out)

    def apply(self, data_1: TensorOrArray, data_2: TensorOrArray) -> TensorOrArray:
        if isinstance(data_1, np.ndarray) and isinstance(data_2, np.ndarray):
            return super().apply(data_1, data_2)
        elif isinstance(data_1, torch.Tensor) and isinstance(data_2, torch.Tensor):
            return self.apply_torch(data_1, data_2)
        else:
            raise ValueError(f"Arguments can only be both numpy arrays or torch tensors but were {data_1} and {data_2}")



import warnings

import numpy as np

from threadpoolctl import threadpool_limits

from ..base import Estimator, Model
from .parallel import handle_n_jobs, joining
from .platform import handle_progress_bar


def _imap_wrapper(args):
    with threadpool_limits(limits=1, user_api='blas'):
        i, fun, arguments = args
        result = fun(*arguments)
        return i, result


class LaggedModelValidation(Model):
    def __init__(self, estimates=None, estimates_conf=None, predictions=None, predictions_conf=None, lagtimes=None,
                 conf=None):
        self._estimates = estimates
        self._estimates_conf = estimates_conf
        self._predictions = predictions
        self._predictions_conf = predictions_conf
        self._lagtimes = lagtimes
        self._conf = conf

    @property
    def lagtimes(self):
        return self._lagtimes

    @property
    def estimates(self):
        """ Returns estimates at different lagtimes

        Returns
        -------
        Y : ndarray(T, n)
            each row contains the n observables computed at one of the T lagtimes.
        """
        return self._estimates

    @property
    def estimates_conf(self):
        """ Returns the confidence intervals of the estimates at different
        lagtimes (if available).

        If not available, returns None.

        Returns
        -------
        estimates_conf : ndarray(T, n, k)
            each row contains the lower confidence bound of n observables
            computed at one of the T lag times.

        """
        return self._estimates_conf

    @property
    def predictions(self):
        """ Returns tested model predictions at different lagtimes

        Returns
        -------
        Y : ndarray(T, n)
            each row contains the n observables predicted at one of the T lag
            times by the tested model.

        """
        return self._predictions

    @property
    def predictions_conf(self):
        """ Returns the confidence intervals of the estimates at different
        lagtimes (if available)

        If not available, returns None.

        Returns
        -------
        L : ndarray(T, n)
            each row contains the lower confidence bound of n observables
            computed at one of the T lag times.

        R : ndarray(T, n)
            each row contains the upper confidence bound of n observables
            computed at one of the T lag times.

        """
        return self._predictions_conf

    @property
    def nsets(self):
        return self._predictions.shape[1]

    @property
    def has_errors(self):
        return self.predictions_conf is not None

    @property
    def err_est(self):
        return self.estimates_conf is not None

    @property
    def conf(self):
        return self._conf


class LaggedModelValidator(Estimator):
    r""" Validates a model estimated at lag time tau by testing its predictions
    for longer lag times

    Parameters
    ----------
    test_model : Model
        Model to be tested

    test_estimator : Estimator
        Parametrized Estimator that has produced the model

    mlags : int or int-array, default=10
        multiples of lag times for testing the Model, e.g. range(10).
        A single int will trigger a range, i.e. mlags=10 maps to
        mlags=range(1, 10). The setting None will choose mlags automatically
        according to the longest available trajectory

    conf : float, default = 0.95
        confidence interval for errors

    err_est : bool, default=False
        if the Estimator is capable of error calculation, will compute
        errors for each tau estimate. This option can be computationally
        expensive.

    """

    def __init__(self, test_model: Model, test_estimator: Estimator, test_model_lagtime: int,
                 mlags, conf=0.95, err_est=False):
        # set model and estimator
        self._test_model = test_model
        import copy
        self.test_estimator = copy.deepcopy(test_estimator)

        self.input_lagtime = test_model_lagtime

        # set conf and error handling
        self.conf = conf
        self.has_errors = hasattr(test_model, 'samples')
        self._set_mlags(mlags)

        self.err_est = err_est
        if err_est and not self.has_errors:
            raise ValueError('Requested errors on the estimated models, '
                             'but the model is not able to calculate errors at all')

        super(LaggedModelValidator, self).__init__()

    def _set_mlags(self, mlags):
        # set mlags, we do it in fit, so we can obtain the maximum possible lagtime from the trajectory data.
        from numbers import Integral
        if isinstance(mlags, Integral):
            mlags = np.arange(mlags)
        mlags = np.asarray(mlags, dtype=int)

        mlags = np.atleast_1d(mlags)
        if (mlags < 0).any():
            raise ValueError('multiples of lagtimes have to be greater equal zero.')
        self.mlags = mlags

    def _effective_mlags(self, data, lagtime):
        if not isinstance(data, list):
            data = [data]
        maxlength = np.max([len(x) for x in data])
        maxmlag = int(np.floor(maxlength / lagtime))
        mlags = np.copy(self.mlags)

        if np.any(mlags > maxmlag):
            mlags = mlags[np.where(mlags <= maxmlag)]
            warnings.warn('Dropped lag times exceeding data lengths')

        return mlags

    def fit(self, data, n_jobs=None, progress=None, estimate_model_for_lag=None, **kw):
        assert estimate_model_for_lag is not None
        n_jobs = handle_n_jobs(n_jobs)
        progress = handle_progress_bar(progress)
        # set lag times

        mlags = self._effective_mlags(data, self.input_lagtime)
        lags = mlags * self.input_lagtime

        predictions = []
        predictions_conf = []
        estimates = []
        estimates_conf = []
        estimated_models = []

        # do we have zero lag? this must be treated separately
        include0 = mlags[0] == 0
        if include0:
            lags_for_estimation = lags[1:]
        else:
            lags_for_estimation = lags

        # estimate models at multiple of input lag time.
        if n_jobs == 1:
            for lag in progress(lags_for_estimation, total=len(lags_for_estimation), leave=False):
                estimated_models.append(estimate_model_for_lag(self.test_estimator, self._test_model, data, lag))
        else:
            from multiprocessing import get_context
            fun = estimate_model_for_lag
            args = [(i, fun, (self.test_estimator, self._test_model, data, lag))
                    for i, lag in enumerate(lags_for_estimation)]
            estimated_models = [None for _ in range(len(args))]
            with joining(get_context("spawn").Pool(processes=n_jobs)) as pool:
                for result in progress(pool.imap_unordered(_imap_wrapper, args),
                                       total=len(lags_for_estimation), leave=False):
                    estimated_models[result[0]] = result[1]

        if include0:
            estimated_models = [None] + estimated_models

        for mlag in mlags:
            # make a prediction using the test model
            predictions.append(self._compute_observables(self._test_model, mlag=mlag))
            # compute prediction errors if we can
            if self.has_errors:
                l, r = self._compute_observables_conf(self._test_model, mlag=mlag)
                predictions_conf.append((l, r))

        for model in estimated_models:
            # evaluate the estimate at lagtime*mlag
            estimates.append(self._compute_observables(model, 1))
            if self.has_errors and self.err_est:
                l, r = self._compute_observables_conf(model, 1)
                estimates_conf.append((l, r))

        # build arrays
        estimates = np.array(estimates)
        predictions = np.array(predictions)
        if self.has_errors:
            predictions_conf = np.array(predictions_conf)
            predictions_conf = predictions_conf.transpose(1, 0, 2, 3)
        else:
            predictions_conf = None
        if self.has_errors and self.err_est:
            estimates_conf = np.array(estimates_conf)
        else:
            estimates_conf = None

        self._model = LaggedModelValidation(
            estimates=estimates, estimates_conf=estimates_conf,
            predictions=predictions, predictions_conf=predictions_conf, lagtimes=lags, conf=self.conf)

        return self

    def _compute_observables(self, model, mlag):
        """Compute observables for given model

        Parameters
        ----------
        model : Model
            model to compute observable for.

        mlag : int
            if 1, just compute the observable for given model. If not 1, use
            model to predict result at multiple of given model lagtime.

        Returns
        -------
        Y : ndarray
            array with results

        """
        raise NotImplementedError()

    def _compute_observables_conf(self, model, mlag, conf=0.95):
        """Compute confidence interval for observables for given model

        Parameters
        ----------
        model : Model
            model to compute observable for. model can be None if mlag=0.
            This scenario must be handled.

        mlag : int, default=1
            if 1, just compute the observable for given model. If not 1, use
            model to predict result at multiple of given model lagtime.

        Returns
        -------
        L : ndarray
            array with lower confidence bounds
        R : ndarray
            array with upper confidence bounds

        """
        raise NotImplementedError()



"""
Created on Jul 23, 2014

@author: noe
"""

import numpy as np
import scipy
import random


def determine_lengths(dtrajs):
    r"""
    Determines the lengths of all trajectories

    Parameters
    ----------
    dtrajs : list of int-arrays
        discrete trajectories
    """
    if isinstance(dtrajs[0], int):
        return len(dtrajs) * np.ones((1,))
    lengths = np.zeros((len(dtrajs)))
    for i in range(len(dtrajs)):
        lengths[i] = len(dtrajs[i])
    return lengths


def bootstrap_trajectories(trajs, correlation_length):
    """
    Generates a randomly resampled count matrix given the input coordinates.

    See API function for full documentation.
    """
    from scipy.stats import rv_discrete
    # if we have just one trajectory, put it into a one-element list:
    if isinstance(trajs[0], (int, int, float)):
        trajs = [trajs]
    ntraj = len(trajs)

    # determine correlation length to be used
    lengths = determine_lengths(trajs)
    Ltot = np.sum(lengths)
    Lmax = np.max(lengths)
    if correlation_length < 1:
        correlation_length = Lmax

        # assign probabilites to select trajectories
    w_trajs = np.zeros((len(trajs)))
    for i in range(ntraj):
        w_trajs[i] = len(trajs[i])
    w_trajs /= np.sum(w_trajs)  # normalize to sum 1.0
    distrib_trajs = rv_discrete(values=(list(range(ntraj)), w_trajs))

    # generate subtrajectories
    Laccum = 0
    subs = []
    while (Laccum < Ltot):
        # pick a random trajectory
        itraj = distrib_trajs.rvs()
        # pick a starting frame
        t0 = random.randint(0, max(1, len(trajs[itraj]) - correlation_length))
        t1 = min(len(trajs[itraj]), t0 + correlation_length)
        # add new subtrajectory
        subs.append(trajs[itraj][t0:t1])
        # increment available states
        Laccum += (t1 - t0)

    # and return
    return subs


def bootstrap_counts_singletraj(dtraj, lagtime, n):
    """
    Samples n counts at the given lagtime from the given trajectory
    """
    # check if length is sufficient
    L = len(dtraj)
    if lagtime > L:
        raise ValueError(f'Cannot sample counts with lagtime {lagtime} from a trajectory with length {L}.')
    # sample
    I = np.random.randint(0, L - lagtime - 1, size=n)
    J = I + lagtime

    # return state pairs
    return dtraj[I], dtraj[J]


def bootstrap_counts(dtrajs, lagtime, corrlength=None):
    """
    Generates a randomly resampled count matrix given the input coordinates.

    See API function for full documentation.
    """
    from scipy.stats import rv_discrete
    # if we have just one trajectory, put it into a one-element list:
    if not isinstance(dtrajs, list):
        dtrajs = [dtrajs]
    ntraj = len(dtrajs)

    # can we do the estimate?
    lengths = determine_lengths(dtrajs)
    Lmax = np.max(lengths)
    Ltot = np.sum(lengths)
    if lagtime >= Lmax:
        raise ValueError('Cannot estimate count matrix: lag time '
                         + str(lagtime) + ' is longer than the longest trajectory length ' + str(Lmax))

    # how many counts can we sample?
    if corrlength is None:
        corrlength = lagtime
    nsample = int(Ltot / corrlength)

    # determine number of states n
    from deeptime.markov import number_of_states
    n = number_of_states(dtrajs)

    # assigning trajectory sampling weights
    w_trajs = np.maximum(0.0, lengths - lagtime)
    w_trajs /= np.sum(w_trajs)  # normalize to sum 1.0
    distrib_trajs = rv_discrete(values=(list(range(ntraj)), w_trajs))
    # sample number of counts from each trajectory
    n_from_traj = np.bincount(distrib_trajs.rvs(size=nsample), minlength=ntraj)

    # for each trajectory, sample counts and stack them
    rows = np.zeros((nsample,))
    cols = np.zeros((nsample,))
    ones = np.ones((nsample,))
    ncur = 0
    for i in range(len(n_from_traj)):
        if n_from_traj[i] > 0:
            (r, c) = bootstrap_counts_singletraj(dtrajs[i], lagtime, n_from_traj[i])
            rows[ncur:ncur + n_from_traj[i]] = r
            cols[ncur:ncur + n_from_traj[i]] = c
            ncur += n_from_traj[i]
    # sum over counts
    Csparse = scipy.sparse.coo_matrix((ones, (rows, cols)), shape=(n, n))

    return Csparse.tocsr()



r"""This module provides matrix-decomposition based functions for the
analysis of stochastic matrices

.. moduleauthor:: B.Trendelkamp-Schroer <benjamin DOT trendelkamp-schroer AT fu-berlin DOT de>
.. moduleauthor:: M.Hoffmann
"""

import numpy as np
import scipy.sparse as sparse
import scipy.linalg as la
import scipy.sparse.linalg as sla
import warnings

from deeptime.util.exceptions import ImaginaryEigenValueWarning, SpectralWarning

from ._stationary_vector import stationary_distribution
from ._assessment import is_reversible


def _symmetrize(T, mu=None):
    if mu is None:
        mu = stationary_distribution(T)
    if np.any(mu <= 0):
        raise ValueError('Cannot symmetrize transition matrix')
    """ symmetrize T """
    smu = np.sqrt(mu)
    if sparse.issparse(T):
        D = sparse.diags(smu, 0)
        Dinv = sparse.diags(1.0 / smu, 0)
        S = (D.dot(T)).dot(Dinv)
    else:
        S = smu[:, None] * T / smu
    return smu, S


def eigenvalues(T, k=None, ncv=None, reversible=False, mu=None):
    r"""Compute the eigenvalues of a sparse transition matrix.

    Parameters
    ----------
    T : (M, M) scipy.sparse matrix or ndarray
        Transition matrix
    k : int, optional
        Number of eigenvalues to compute.
    ncv : int, optional
        The number of Lanczos vectors generated, `ncv` must be greater than k;
        it is recommended that ncv > 2*k
    reversible : bool, optional
        Indicate that transition matrix is reversible
    mu : (M,) ndarray, optional
        Stationary distribution of T

    Returns
    -------
    v : (k,) ndarray
        Eigenvalues of T

    Notes
    -----
    The first k eigenvalues of largest magnitude are computed.

    If reversible=True the the eigenvalues of the similar symmetric
    matrix `\sqrt(\mu_i / \mu_j) p_{ij}` will be computed.

    The precomputed stationary distribution will only be used if
    reversible=True.

    """
    is_sparse = sparse.issparse(T)
    if is_sparse and k is None:
        raise ValueError("Number of eigenvalues required for decomposition of sparse matrix")
    else:
        if reversible:
            try:
                v = eigenvalues_rev(T, k, ncv=ncv, mu=mu)
            except ValueError:  # use fallback code, but cast to real
                if is_sparse:
                    v = sla.eigs(T, k=k, which='LM', return_eigenvectors=False, ncv=ncv).real
                else:
                    v = la.eigvals(T).real
        else:
            # nonreversible
            if is_sparse:
                v = sla.eigs(T, k=k, which='LM', return_eigenvectors=False, ncv=ncv)
            else:
                v = la.eigvals(T)

    ind = np.argsort(np.abs(v))[::-1][:k]  # top k
    return v[ind]


def eigenvalues_rev(T, k, ncv=None, mu=None):
    r"""Compute the eigenvalues of a reversible, sparse transition matrix.

    Parameters
    ----------
    T : (M, M) scipy.sparse matrix
        Transition matrix
    k : int
        Number of eigenvalues to compute.
    ncv : int, optional
        The number of Lanczos vectors generated, `ncv` must be greater than k;
        it is recommended that ncv > 2*k
    mu : (M,) ndarray, optional
        Stationary distribution of T

    Returns
    -------
    v : (k,) ndarray
        Eigenvalues of T

    Raises
    ------
    ValueError
        If stationary distribution is nonpositive.

    Notes
    -----
    The first k eigenvalues of largest magnitude are computed.

    """

    _, S = _symmetrize(T, mu=mu)
    """Compute eigenvalues using a solver for symmetric/hermititan eigenproblems"""
    if sparse.issparse(T):
        evals = sla.eigsh(S, k=k, ncv=ncv, which='LM', return_eigenvectors=False)
    else:
        evals = la.eigvalsh(S)
    return evals


def eigenvectors(T, k=None, right=True, reversible=False, mu=None, ncv=None):
    r"""Compute eigenvectors of transition matrix.

    Parameters
    ----------
    T : (d, d) ndarray
        Transition matrix (stochastic matrix)
    k : int or tuple of ints, optional
        Compute the first k eigenvalues of T
    right : bool, optional
        If right=True compute right eigenvectors, left eigenvectors
        otherwise
    reversible : bool, optional
        Indicate that transition matrix is reversible
    mu : (d,) ndarray, optional
        Stationary distribution of T

    Returns
    -------
    eigvec : (d, n) ndarray
        The eigenvectors of T ordered with decreasing absolute value
        of the corresponding eigenvalue. If k is None then n=d, if k
        is int then n=k otherwise n is the length of the given tuple
        of eigenvector indices

    Notes
    -----
    Eigenvectors are computed using the numpy.linalg interface for the
    corresponding LAPACK routines.

    If reversible=True the the eigenvectors of the similar symmetric
    matrix `\sqrt(\mu_i / \mu_j) p_{ij}` will be used to compute the
    eigenvectors of T.

    The precomputed stationary distribution will only be used if
    reversible=True.

    """
    is_sparse = sparse.issparse(T)

    if is_sparse and k is None:
        raise ValueError("Number of eigenvectors required for decomposition of sparse matrix")

    if reversible:
        smu, S = _symmetrize(T, mu=mu)
        if is_sparse:
            val, vecs = sla.eigsh(S, k=k, ncv=ncv, which='LM', return_eigenvectors=True)
        else:
            val, vecs = la.eigh(S)
        """Sort eigenvectors"""
        perm = np.argsort(np.abs(val))[::-1][:k]
        eigvec = vecs[:, perm]
        if right:
            eigvec /= smu[:, None]
        else:
            eigvec *= smu[:, None]
    else:
        if not is_sparse:
            val, vecs = la.eig(T, left=not right, right=right)
        else:
            val, vecs = sla.eigs(T if right else T.transpose(), k=k, which='LM', ncv=ncv)
        perm = np.argsort(np.abs(val))[::-1][:k]
        eigvec = vecs[:, perm]

    return eigvec


def rdl_decomposition(T, k=None, ncv=None, reversible=False, norm='standard', mu=None):
    r"""Compute the decomposition into left and right eigenvectors.

    Parameters
    ----------
    T : (M, M) ndarray
        Transition matrix
    k : int (optional)
        Number of eigenvector/eigenvalue pairs
    norm: {'standard', 'reversible', 'auto'}
        standard: (L'R) = Id, L[:,0] is a probability distribution,
            the stationary distribution mu of T. Right eigenvectors
            R have a 2-norm of 1.
        reversible: R and L are related via L=L[:,0]*R.
        auto: will be reversible if T is reversible, otherwise standard
    reversible : bool, optional
        Indicate that transition matrix is reversible
    mu : (d,) ndarray, optional
        Stationary distribution of T

    Returns
    -------
    R : (M, M) ndarray
        The normalized (with respect to L) right eigenvectors, such that the
        column R[:,i] is the right eigenvector corresponding to the eigenvalue
        w[i], dot(T,R[:,i])=w[i]*R[:,i]
    D : (M, M) ndarray
        A diagonal matrix containing the eigenvalues, each repeated
        according to its multiplicity
    L : (M, M) ndarray
        The normalized (with respect to `R`) left eigenvectors, such that the
        row ``L[i, :]`` is the left eigenvector corresponding to the eigenvalue
        ``w[i]``, ``dot(L[i, :], T)``=``w[i]*L[i, :]``

    Notes
    -----
    If reversible=True the the eigenvalues and eigenvectors of the
    similar symmetric matrix `\sqrt(\mu_i / \mu_j) p_{ij}` will be
    used to compute the eigenvalues and eigenvectors of T.

    The precomputed stationary distribution will only be used if
    reversible=True.

    """
    if sparse.issparse(T) and k is None:
        raise ValueError("Number of eigenvectors required for decomposition of sparse matrix")
    # auto-set norm
    if norm == 'auto':
        if is_reversible(T):
            norm = 'reversible'
        else:
            norm = 'standard'

    if reversible:
        R, D, L = rdl_decomposition_rev(T, norm=norm, mu=mu, k=k, ncv=ncv)
    else:
        R, D, L = rdl_decomposition_nrev(T, norm=norm, k=k, ncv=ncv)

    if reversible or norm == 'reversible':
        D = D.real

    return R[:, :k], D[:k, :k], L[:k, :]


def rdl_decomposition_nrev(T, norm='standard', k=None, ncv=None):
    r"""Decomposition into left and right eigenvectors.

    Parameters
    ----------
    T : (M, M) ndarray
        Transition matrix
    norm: {'standard', 'reversible'}
        standard: (L'R) = Id, L[:,0] is a probability distribution,
            the stationary distribution mu of T. Right eigenvectors
            R have a 2-norm of 1
        reversible: R and L are related via L=L[:,0]*R

    Returns
    -------
    R : (M, M) ndarray
        The normalized (with respect to L) right eigenvectors, such that the
        column R[:,i] is the right eigenvector corresponding to the eigenvalue
        w[i], dot(T,R[:,i])=w[i]*R[:,i]
    D : (M, M) ndarray
        A diagonal matrix containing the eigenvalues, each repeated
        according to its multiplicity
    L : (M, M) ndarray
        The normalized (with respect to `R`) left eigenvectors, such that the
        row ``L[i, :]`` is the left eigenvector corresponding to the eigenvalue
        ``w[i]``, ``dot(L[i, :], T)``=``w[i]*L[i, :]``


    """
    is_sparse = sparse.issparse(T)
    d = T.shape[0]
    if is_sparse:
        v, R = sla.eigs(T, k=k, which='LM', ncv=ncv)
    else:
        v, R = la.eig(T)

    """Sort by decreasing magnitude of eigenvalue"""
    ind = np.argsort(np.abs(v))[::-1]
    v = v[ind]
    R = R[:, ind]

    """Diagonal matrix containing eigenvalues"""
    D = np.diag(v)

    # Standard norm: Euclidean norm is 1 for r and LR = I.
    if norm == 'standard':
        if is_sparse:
            r, L = sla.eigs(T.transpose(), k=k, which='LM', ncv=ncv)
            ind = np.argsort(np.abs(r))[::-1]
            L = L[:, ind]
        else:
            L = la.solve(np.transpose(R), np.eye(d))

        """l1- normalization of L[:, 0]"""
        R[:, 0] = R[:, 0] * np.sum(L[:, 0])
        L[:, 0] = L[:, 0] / np.sum(L[:, 0])

        if is_sparse:  # in dense case we already got this
            """Standard normalization L'R=Id"""
            ov = np.diag(np.dot(np.transpose(L), R))
            R = R / ov[np.newaxis, :]

        return R, D, np.transpose(L)

    # Reversible norm:
    elif norm == 'reversible':
        if is_sparse:
            mu = stationary_distribution(T)
        else:
            b = np.zeros(d)
            b[0] = 1.0

            A = np.transpose(R)
            nu = la.solve(A, b)
            mu = nu / np.sum(nu)

        """Ensure that R[:,0] is positive"""
        R[:, 0] = R[:, 0] / np.sign(R[0, 0])

        """Use mu to connect L and R"""
        L = mu[:, np.newaxis] * R

        """Compute overlap"""
        s = np.diag(np.dot(np.transpose(L), R))

        """Renormalize left-and right eigenvectors to ensure L'R=Id"""
        R = R / np.sqrt(s[np.newaxis, :])
        L = L / np.sqrt(s[np.newaxis, :])

        return R, D, np.transpose(L)

    else:
        raise ValueError("Keyword 'norm' has to be either 'standard' or 'reversible'")


def rdl_decomposition_rev(T, norm='reversible', mu=None, ncv=None, k=None):
    r"""Decomposition into left and right eigenvectors for reversible
    transition matrices.

    Parameters
    ----------
    T : (M, M) ndarray
        Transition matrix
    norm: {'standard', 'reversible'}
        standard: (L'R) = Id, L[:,0] is a probability distribution,
            the stationary distribution mu of T. Right eigenvectors
            R have a 2-norm of 1.
        reversible: R and L are related via L=L[:,0]*R.
    mu : (M,) ndarray, optional
        Stationary distribution of T

    Returns
    -------
    R : (M, M) ndarray
        The normalized (with respect to L) right eigenvectors, such that the
        column R[:,i] is the right eigenvector corresponding to the eigenvalue
        w[i], dot(T,R[:,i])=w[i]*R[:,i]
    D : (M, M) ndarray
        A diagonal matrix containing the eigenvalues, each repeated
        according to its multiplicity
    L : (M, M) ndarray
        The normalized (with respect to `R`) left eigenvectors, such that the
        row ``L[i, :]`` is the left eigenvector corresponding to the eigenvalue
        ``w[i]``, ``dot(L[i, :], T)``=``w[i]*L[i, :]``

    Notes
    -----
    The eigenvalues and eigenvectors of the similar symmetric matrix
    `\sqrt(\mu_i / \mu_j) p_{ij}` will be used to compute the
    eigenvalues and eigenvectors of T.

    The stationay distribution will be computed if no precomputed stationary
    distribution is given.

    """
    """ symmetrize T """
    is_sparse = sparse.issparse(T)
    smu, S = _symmetrize(T, mu)
    if is_sparse:
        val, eigvec = sla.eigsh(S, k=k, ncv=ncv, which='LM', return_eigenvectors=True)
    else:
        val, eigvec = la.eigh(S)

    """Sort eigenvalues and eigenvectors"""
    perm = np.argsort(np.abs(val))[::-1]
    val = val[perm]
    eigvec = eigvec[:, perm]

    """Diagonal matrix of eigenvalues"""
    D = np.diag(val)

    """Right and left eigenvectors"""
    R = eigvec / smu[:, np.newaxis]
    L = eigvec * smu[:, np.newaxis]

    """Ensure that R[:,0] is positive and unity"""
    tmp = R[0, 0]
    R[:, 0] = R[:, 0] / tmp

    """Ensure that L[:, 0] is probability vector"""
    L[:, 0] = L[:, 0] * tmp

    if norm == 'reversible':
        return R, D, L.T
    elif norm == 'standard':
        """Standard l2-norm of right eigenvectors"""
        w = np.diag(np.dot(R.T, R))
        sw = np.sqrt(w)
        """Don't change normalization of eigenvectors for dominant eigenvalue"""
        sw[0] = 1.0

        R = R / sw[np.newaxis, :]
        L = L * sw[np.newaxis, :]
        return R, D, L.T
    else:
        raise ValueError("Keyword 'norm' has to be either 'standard' or 'reversible'")


def timescales(T, tau=1, k=None, ncv=None, reversible=False, mu=None):
    r"""Compute implied time scales of given transition matrix

    Parameters
    ----------
    T : (M, M) ndarray
        Transition matrix
    tau : int, optional
        lag time
    k : int, optional
        Number of time scales
    reversible : bool, optional
        Indicate that transition matirx is reversible
    mu : (M,) ndarray, optional
        Stationary distribution of T

    Returns
    -------
    ts : (N,) ndarray
        Implied time scales of the transition matrix.
        If k=None then N=M else N=k

    Notes
    -----
    If reversible=True the the eigenvalues of the similar symmetric
    matrix `\sqrt(\mu_i / \mu_j) p_{ij}` will be computed.

    The precomputed stationary distribution will only be used if
    reversible=True.

    """
    if sparse.issparse(T) and k is None:
        raise ValueError("Number of time scales required for decomposition of sparse matrix")
    values = eigenvalues(T, reversible=reversible, mu=mu, k=k, ncv=ncv)
    return timescales_from_eigenvalues(values, tau)


def timescales_from_eigenvalues(evals, tau=1):
    r"""Compute implied time scales from given eigenvalues

    Parameters
    ----------
    evals : eigenvalues
    tau : lag time

    Returns
    -------
    ts : ndarray
        The implied time scales to the given eigenvalues, in the same order.

    """

    """Check for dominant eigenvalues with large imaginary part"""

    if not np.allclose(evals.imag, 0.0):
        warnings.warn('Using eigenvalues with non-zero imaginary part', ImaginaryEigenValueWarning)

    """Check for multiple eigenvalues of magnitude one"""
    ind_abs_one = np.isclose(np.abs(evals), 1.0, rtol=0.0, atol=1e-14)
    if sum(ind_abs_one) > 1:
        warnings.warn('Multiple eigenvalues with magnitude one.', SpectralWarning)

    """Compute implied time scales"""
    ts = np.zeros(len(evals))

    """Eigenvalues of magnitude one imply infinite timescale"""
    ts[ind_abs_one] = np.inf

    """All other eigenvalues give rise to finite timescales"""
    ts[np.logical_not(ind_abs_one)] = \
        -1.0 * tau / np.log(np.abs(evals[np.logical_not(ind_abs_one)]))
    return ts



"""
    Sparse assessment module of msm analysis package
"""

import numpy as np
import scipy.sparse as sparse

from deeptime.numeric import allclose_sparse
from ._stationary_vector import stationary_distribution


def is_transition_matrix(T, tol):
    """
    True if T is a transition matrix

    Parameters
    ----------
    T : scipy.sparse matrix
        Matrix to check
    tol : float
        tolerance to check with

    Returns
    -------
    Truth value: bool
        True, if T is positive and normed
        False, otherwise

    """
    if T.ndim != 2 or T.shape[0] != T.shape[1]:
        return False

    if sparse.issparse(T):
        T = T.tocsr()  # compressed sparse row for fast row slicing
        values = T.data  # non-zero entries of T
    else:
        values = T

    is_positive = np.allclose(values, np.abs(values), rtol=tol)
    is_normed = np.allclose(T.sum(axis=1), 1.0, rtol=tol)

    return is_positive and is_normed


def is_rate_matrix(K, tol):
    r""" True if K is a rate matrix

    Parameters
    ----------
    K : scipy.sparse matrix
        Matrix to check
    tol : float
        tolerance to check with

    Returns
    -------
    Truth value : bool
        True, if K negated diagonal is positive and row sums up to zero.
        False, otherwise
    """
    if sparse.issparse(K):
        K = K.tocsr()

    # check rows sum up to zero.
    row_sum = K.sum(axis=1)
    sum_eq_zero = np.allclose(row_sum, 0., atol=tol)

    R = K - K.diagonal()
    if sparse.issparse(R):
        R = R.values  # extract nonzero entries
    off_diagonal_positive = np.allclose(R, np.abs(R), rtol=0, atol=tol)

    return off_diagonal_positive and sum_eq_zero


def is_reversible(T, mu=None, tol=1e-15):
    r"""
    checks whether T is reversible in terms of given stationary distribution.
    If no distribution is given, it will be calculated out of T.

    performs follwing check:
    :math:`\pi_i P_{ij} = \pi_j P_{ji}
    Parameters
    ----------
    T : scipy.sparse matrix
        Transition matrix
    mu : numpy.ndarray vector
        stationary distribution
    tol : float
        tolerance to check with

    Returns
    -------
    Truth value : bool
        True, if T is a stochastic matrix
        False, otherwise
    """
    if not is_transition_matrix(T, tol):
        raise ValueError("given matrix is not a valid transition matrix.")

    if sparse.issparse(T):
        T = T.tocsr()

    if mu is None:
        mu = stationary_distribution(T)

    if sparse.issparse(T):
        prod = sparse.construct.diags(mu) * T
    else:
        prod = mu[:, None] * T

    if sparse.issparse(T):
        return allclose_sparse(prod, prod.transpose(), rtol=tol)
    else:
        return np.allclose(prod, prod.transpose(), rtol=tol)


def is_connected(T, directed=True):
    r"""Check connectivity of the transition matrix.

    Return true, if the input matrix is completely connected,
    effectively checking if the number of connected components equals one.

    Parameters
    ----------
    T : scipy.sparse matrix
        Transition matrix
    directed : bool, optional
       Whether to compute connected components for a directed  or
       undirected graph. Default is True.

    Returns
    -------
    connected : boolean, returning true only if T is connected.


    """
    if not sparse.issparse(T):
        T = sparse.csr_matrix(T)
    nc = sparse.csgraph.connected_components(T, directed=directed, connection='strong', return_labels=False)
    return nc == 1


def is_ergodic(T, tol):
    """
    checks if T is 'ergodic'

    Parameters
    ----------
    T : scipy.sparse matrix
        Transition matrix
    tol : float
        tolerance

    Returns
    -------
    Truth value : bool
        True, if # strongly connected components = 1
        False, otherwise
    """
    if not sparse.issparse(T):
        T = sparse.csr_matrix(T)
    if not is_transition_matrix(T, tol):
        raise ValueError("given matrix is not a valid transition matrix.")
    return is_connected(T, True)



import numpy as np
from scipy.sparse import diags


class BirthDeathChain:
    r"""Birth and death chain.

    A general birth and death chain on a d-dimensional state space
    has the following transition matrix

    .. math::

        p_{ij} = \begin{cases}
            q_i &\text{, if } j=i-1 \text{ and } i>0,\\
            r_i &\text{, if } j=i,\\
            p_i &\text{, if } j=i+1 \text{ and } i < d-1
        \end{cases}

    Parameters
    ----------
    q : array_like
        Annihilation probabilities for transition from i to i-1.
    p : array-like
        Creation probabilities for transition from i to i+1.
    sparse : bool, optional, default=False
        Whether sparse matrices are used.
    """

    def __init__(self, q, p, sparse: bool = False):
        q = np.asarray(q)
        p = np.asarray(p)
        if q[0] != 0.0:
            raise ValueError('Probability q[0] must be zero')
        if p[-1] != 0.0:
            raise ValueError('Probability p[-1] must be zero')
        if not np.all(q + p <= 1.0):
            raise ValueError('Probabilities q+p can not exceed one')
        self.q = q
        self.p = p
        self.r = 1 - self.q - self.p
        self.dim = self.r.shape[0]
        self.sparse = sparse

    @property
    def transition_matrix(self):
        r""" Transition matrix for birth and death chain with given
        creation and anhilation probabilities.

        :getter: Yields the transition matrix.
        :type: (N,N) ndarray
        """
        if not self.sparse:
            P0 = np.diag(self.r, k=0)
            P1 = np.diag(self.p[0:-1], k=1)
            P_1 = np.diag(self.q[1:], k=-1)
            return P0 + P1 + P_1
        else:
            return diags([self.q[1:], self.r, self.p[0:-1]], [-1, 0, 1])

    @property
    def msm(self):
        r""" MarkovStateModel for this birth death chain

        :getter: Yields the MSM.
        :type: deeptime.markov.msm.MarkovStateModel
        """
        from deeptime.markov.msm import MarkovStateModel
        return MarkovStateModel(self.transition_matrix, self.stationary_distribution)

    @property
    def stationary_distribution(self):
        r""" The stationary distribution of the birth-death chain. """
        a = np.zeros(self.dim)
        a[0] = 1.0
        a[1:] = np.cumprod(self.p[0:-1] / self.q[1:])
        mu = a / np.sum(a)
        return mu

    def committor_forward(self, a, b):
        r"""Forward committor for birth-and-death-chain.

        The forward committor is the probability to hit
        state b before hitting state a starting in state x,

        .. math::

            u_x = P_x(T_b<T_a),

        :math:`T_i` is the first arrival time of the chain to state i,

        .. math::

            T_i = \inf ( t>0 \mid X_t=i ).

        Parameters
        ----------
        a : int
            State index
        b : int
            State index

        Returns
        -------
        u : (M,) ndarray
            Vector of committor probabilities.

        """
        u = np.zeros(self.dim)
        g = np.zeros(self.dim - 1)
        g[0] = 1.0
        g[1:] = np.cumprod(self.q[1:-1] / self.p[1:-1])

        """If a and b are equal the event T_b<T_a is impossible
           for any starting state x so that the committor is
           zero everywhere"""
        if a == b:
            return u
        elif a < b:
            """Birth-death chain has to hit a before it can hit b"""
            u[0:a + 1] = 0.0
            """Birth-death chain has to hit b before it can hit a"""
            u[b:] = 1.0
            """Intermediate states are given in terms of sums of g"""
            u[a + 1:b] = np.cumsum(g[a:b])[0:-1] / np.sum(g[a:b])
            return u
        else:
            u[0:b + 1] = 1.0
            u[a:] = 0.0
            u[b + 1:a] = (np.cumsum(g[b:a])[0:-1] / np.sum(g[b:a]))[::-1]
            return u

    def committor_backward(self, a, b):
        r"""Backward committor for birth-and-death-chain.

        The backward committor is the probability for a chain in state x chain to originate from state a instead of
        coming from state b :math:`w_x=P_x(t_a<t_b)`, :math:`t_i` is the last exit time of the chain from state i,
        :math:`t_i = \inf ( t>0 \mid X_{-t} = i )`.

        Parameters
        ----------
        a : int
            State index
        b : int
            State index

        Returns
        -------
        w : (M,) ndarray
            Vector of committor probabilities.

        Notes
        -----
        The birth-death chain is time-reversible

        .. math::

            P(t_a < t_b) = P(T_a < T_b) = 1-P(T_b<T_a),

        therefore we can express the backward comittor probabilities in terms of the
        forward committor probabilities :math:`w=1-u`.
        """
        return 1.0 - self.committor_forward(a, b)

    def flux(self, a, b):
        r"""The flux network for the reaction from
        A=[0,...,a] => B=[b,...,M].

        Parameters
        ----------
        a : int
            State index
        b : int
            State index

        Returns
        -------
        flux : (M, M) ndarray
            Matrix of flux values between pairs of states.

        """
        # if b<=a:
        # raise ValueError("State index b has to be strictly larger than state index a")
        qminus = self.committor_backward(a, b)
        qplus = self.committor_forward(a, b)
        P = self.transition_matrix
        pi = self.stationary_distribution
        from deeptime.markov.tools.flux import flux_matrix
        return flux_matrix(P, pi, qminus, qplus, netflux=False)

    def netflux(self, a, b):
        r"""The netflux network for the reaction from
        A=[0,...,a] => B=[b,...,M].

        Parameters
        ----------
        a : int
            State index
        b : int
            State index

        Returns
        -------
        netflux : (M, M) ndarray
            Matrix of flux values between pairs of states.

        """
        flux = self.flux(a, b)
        from deeptime.markov.tools.flux import to_netflux
        return to_netflux(flux)

    def totalflux(self, a, b):
        r"""The tiotal flux for the reaction A=[0,...,a] => B=[b,...,M].

        Parameters
        ----------
        a : int
            State index
        b : int
            State index

        Returns
        -------
        F : float
            The total flux between reactant and product

        """
        flux = self.flux(a, b)
        A = list(range(a + 1))
        from deeptime.markov.tools.flux import total_flux
        return total_flux(flux, A)

    def rate(self, a, b):
        r""" Yields the total transition rate between state sets `A=[0,...,a]` and `B=[0,...,b]`.

        Parameters
        ----------
        a : int
            State index.
        b : int
            State index.

        Returns
        -------
        kAB : float
            Total transition rate.
        """
        F = self.totalflux(a, b)
        pi = self.stationary_distribution
        qminus = self.committor_backward(a, b)
        kAB = F / (pi * qminus).sum()
        return kAB



from deeptime.util.platform import module_available
if not module_available("torch"):
    raise ValueError("Importing this module is only possible with a working installation of PyTorch.")
del module_available

import numpy as np

from pathlib import Path
from typing import List, Dict, Optional

import torch
import torch.nn as nn


def map_data(data, device=None, dtype=np.float32) -> List[torch.Tensor]:
    r""" Maps data (or list/tuple of data) to torch tensors of a specific dtype and device.

    Parameters
    ----------
    data : ndarray or tensor or list thereof
        The input data.
    device : Device or None
        The device.
    dtype : dtype, default=None
        The dtype.

    Yields
    ------
    mapped_data : list of torch Tensor
        The mapped data.
    """
    with torch.no_grad():
        if not isinstance(data, (list, tuple)):
            data = [data]
        for x in data:
            if isinstance(x, torch.Tensor):
                x = x.to(device=device)
            else:
                x = torch.from_numpy(np.asarray(x, dtype=dtype).copy()).to(device=device)
            yield x


class MLP(nn.Module):
    r""" A multilayer perceptron which can, e.g., be used as a neural network lobe for VAMPNets.

    Parameters
    ----------
    units : list of int
        The units of the fully connected layers.
    nonlinearity : callable, default=torch.nn.ELU
        A callable (like a constructor) which yields an instance of a particular activation function.
    initial_batchnorm : bool, default=True
        Whether to use batch normalization before the data enters the rest of the network.
    output_nonlinearity : callable, default=None
        The output activation/nonlinearity. If the data decomposes into states, it can make sense to use
        an output activation like softmax which produces a probability distribution over said states.
        The callable should take no arguments and produce an object of type :code:`torch.nn.Module`.
    """

    def __init__(self, units: List[int], nonlinearity=nn.ELU, initial_batchnorm: bool = False,
                 output_nonlinearity=None):
        super().__init__()
        if len(units) > 1:
            layers = []
            if initial_batchnorm:
                layers.append(nn.BatchNorm1d(units[0]))
            for fan_in, fan_out in zip(units[:-2], units[1:-1]):
                layers.append(nn.Linear(fan_in, fan_out))
                layers.append(nonlinearity())
            layers.append(nn.Linear(units[-2], units[-1]))
            if output_nonlinearity is not None:
                layers.append(output_nonlinearity())
            self._sequential = nn.Sequential(*layers)
        else:
            self._sequential = nn.Identity()

    def forward(self, inputs):
        return self._sequential(inputs)


class CheckpointManager:
    r""" A checkpoint manager for pytorch models. It can keep track of a metric, save the best model according to
    that metric, and prune too old checkpoints. """

    def __init__(self, output_dir, keep_n_checkpoints: int = 5, best_metric_mode='max'):
        r""" Creates a new checkpoint manager.

        Parameters
        ----------
        output_dir : path_like
            The output directory under which checkpoints are stored. The (according to metric) best checkpoint gets
            assigned the filename "best.ckpt", the others are enumerated as "checkpoint_{step}.ckpt".
        keep_n_checkpoints : int, default=5
            The number of sequential checkpoints to keep. The manager will prune older checkpoints based
            on filename.
        best_metric_mode : str, default='max'
            Whether the smallest or the largest metric is the best. Defaults to 'max', i.e., the largest.
        """
        self._output_dir = Path(output_dir)
        self._output_dir.mkdir(parents=True, exist_ok=True)
        self._keep_n_checkpoints = keep_n_checkpoints
        self._best_metric_value = None
        self._best_metric_mode = best_metric_mode

    @property
    def keep_n_checkpoints(self) -> int:
        r""" The number of checkpoints to keep.

        :getter: Yields the number of checkpoints to keep.
        :setter: Sets the number of checkpoints to keep. Pruning can be called manually or upon next :meth:`step`.
        :type: int
        """
        return self._keep_n_checkpoints

    @keep_n_checkpoints.setter
    def keep_n_checkpoints(self, value: int):
        self._keep_n_checkpoints = value

    @property
    def best_metric_value(self) -> Optional[float]:
        r""" The current best metric value. Can initially be None.

        :getter: The best value.
        :type: float or None
        """
        return self._best_metric_value

    @property
    def best_metric_mode(self) -> str:
        r""" The mode under which the metric is taken into account. Can either be 'max', i.e., larger metric values are
        better, or 'min' for the opposite behavior.

        :getter: Yields the mode.
        :setter: Sets the mode.
        :type: str, one of ('max', 'min')
        """
        return self._best_metric_mode

    @best_metric_mode.setter
    def best_metric_mode(self, value: str):
        assert value in ('min', 'max'), f"Unknown mode {value}, supported are min and max."
        self._best_metric_mode = value

    def step(self, step: int, metric_value: float, models: Dict[str, torch.nn.Module]):
        r""" Records a step in the checkpoint manager. This automatically prunes old checkpoints according to
        :meth:`max_n_checkpoints`.

        Parameters
        ----------
        step : int
            The training step. Should be monotonically increasing during training.
        metric_value : float
            Value of the metric.
        models : dict from model name to torch module
            Dictionary of modules which should be stored in the checkpoint.
        """
        self._make_checkpoint(step, models, self._output_dir / f"checkpoint_{step}.ckpt")
        self.prune_checkpoints()
        if metric_value is not None:
            op = min if self.best_metric_mode == 'min' else max
            if self.best_metric_value is None or (metric_value == op(self.best_metric_value, metric_value)):
                self._best_metric_value = metric_value
                self._make_checkpoint(step, models, self._output_dir / f"best.ckpt")

    @staticmethod
    def _make_checkpoint(step, models: Dict[str, torch.nn.Module], outfile: Path):
        r""" Makes the actual checkpoint. """
        save_dict = {k: v.state_dict() for k, v in models.items()}
        save_dict['step'] = step
        torch.save(save_dict, outfile)
        return outfile

    def prune_checkpoints(self):
        r""" Prunes old checkpoints. """
        checkpoints = list(self._output_dir.glob("*.ckpt"))
        steps = []
        for ckpt in checkpoints:
            fname = str(Path(ckpt).name)
            filtered = "".join(list(c for c in filter(str.isdigit, fname)))
            if len(filtered) > 0:
                steps.append(int(filtered))
        while len(steps) > self.keep_n_checkpoints:
            oldest = min(steps)
            oldest_path = self._output_dir.joinpath(
                "checkpoint_{}.ckpt".format(oldest)
            )
            oldest_path.unlink()
            steps.remove(oldest)


class Stats:
    r""" Object that collects training statistics in a certain group. This can be used to track, e.g., loss values
    and validation values.

    >>> stats = Stats(group='train', items=['loss1', 'loss2'])

    Adding some artificial data (first list element belongs to loss1, second to loss2).
    This could be recording the loss during an epoch of training.
    >>> stats.add([torch.tensor(1.), torch.tensor(1.)])
    >>> stats.add([torch.tensor(2.), torch.tensor(2.)])
    >>> stats.add([torch.tensor(3.), torch.tensor(3.)])

    After the epoch, the mean value of the collected stats can be written using a tensorboard summary writer:

    .. code-block:: python
        from torch.utils.tensorboard import SummaryWriter
        writer = SummaryWriter(log_dir="path/to/train/logdir")
        stats.write(writer, global_step=step)  # this also clears the stats for the next round of collecting data.
    """

    def __init__(self, group: str, items: List[str]):
        r""" Instantiates a new stats object.

        Parameters
        ----------
        group : str
            The group this stats object belongs to.
        items : list of str
            List of strings
        """
        self._stats = []
        self._group = group
        self._items = items

    def add(self, data: List[torch.Tensor]):
        r""" Adds data to the statistics.

        Parameters
        ----------
        data : list of torch tensors
            Adds a list of tensors. Must be of same length as the number of items that are tracked in this object.
        """
        if len(data) != len(self._items):
            raise ValueError("Incompatible stats")
        self._stats.append(torch.stack([x.detach() for x in data]))

    @property
    def items(self) -> List[str]:
        r""" The items that are tracked by this object.

        :getter: Yields the items.
        :type: list of str
        """
        return self._items

    @property
    def group(self) -> str:
        r""" The group that this object belongs to (e.g., validation or train).

        :getter: Yields the group.
        :type: str
        """
        return self._group

    @property
    def samples(self):
        r""" Property to access the currently stored statistics.

        :getter: Gets the statistics.
        :type: (n_items, n_data) ndarray
        """
        return torch.stack(self._stats).cpu().numpy()

    def write(self, writer, global_step: int = None,
              walltime: float = None, clear: bool = True):
        r"""Writes the current statistics using a tensorboard SummaryWriter or an :class:`OutputHandler`.

        Parameters
        ----------
        writer : SummaryWriter
            A tensorboard summary writer used to write statistics.
        global_step : int, optional, default=None
            Optionally the global step value to record.
        walltime: float, optional, default=None
            Optionally the walltime to record.
        clear : bool, default=True
            Whether to clear the statistics, see also :meth:`clear`.
        """
        stats = torch.stack(self._stats)
        mean_stats = stats.mean(dim=0)
        for ix, item in enumerate(self._items):
            name = self._group + "/" + item
            value = torch.mean(mean_stats[ix]).cpu().numpy()
            writer.add_scalar(name, value, global_step=global_step, walltime=walltime)
        if clear:
            self.clear()

    def clear(self):
        r""" Empties the statistics. This is default behavior if statistics are written to a summary file, but
        sometimes it can be useful to track statistics for some more time and eventually clear it manually. """
        self._stats.clear()



from ._version import get_versions

__version__ = get_versions()['version']
del get_versions

from . import util
from . import numeric
from . import data
from . import basis
from . import kernels
from . import sindy
from . import clustering
from . import covariance
from . import decomposition
from . import markov


def capi_includes(inc_clustering: bool = False, inc_markov: bool = False, inc_markov_hmm: bool = False,
                  inc_data: bool = False):
    import os
    import sys
    module_path = sys.modules['deeptime'].__path__[0]
    includes = [os.path.join(module_path, 'src', 'include')]  # common headers
    if inc_clustering:
        includes.append(os.path.join(module_path, 'clustering', 'include'))
    if inc_markov:
        includes.append(os.path.join(module_path, 'markov', '_bindings', 'include'))
    if inc_markov_hmm:
        includes.append(os.path.join(module_path, 'markov', 'hmm', '_bindings', 'include'))
    if inc_data:
        includes.append(os.path.join(module_path, 'data', 'include'))
    return includes



from typing import List

import numpy as np

from deeptime.util.types import ensure_dtraj_list


def compute_index_states(dtrajs, subset=None) -> List[np.ndarray]:
    """Generates a trajectory/time indices for the given list of states

    Parameters
    ----------
    dtrajs : array_like or list of array_like
        Discretized trajectory or list of discretized trajectories. Negative elements will be ignored
    subset : ndarray((n)), optional, default = None
        array of states to be indexed. By default all states in dtrajs will be used

    Returns
    -------
    indices : List[np.ndarray]
        For each state, all trajectory and time indices where this state occurs.
        Each matrix has a number of rows equal to the number of occurrences of the corresponding state,
        with rows consisting of a tuple (i, t), where i is the index of the trajectory and t is the time index
        within the trajectory.

    """
    # check input
    from .. import _markov_bindings as bd
    dtrajs = ensure_dtraj_list(dtrajs)
    return bd.sample.index_states(dtrajs, subset)


################################################################################
# sampling from state indices
################################################################################


def indices_by_sequence(indices: List[np.ndarray], sequence):
    r"""Samples trajectory/time indices according to the given sequence of states.

    Notes
    -----
    Returns -1 indices for states that are not observed in the current sample.

    Parameters
    ----------
    indices : List[np.ndarray]
        For each state, all trajectory and time indices where this state occurs.
        Each matrix has a number of rows equal to the number of occurrences of the corresponding state,
        with rows consisting of a tuple (i, t), where i is the index of the trajectory and t is the time index
        within the trajectory.
    sequence : array_like of integers
        A sequence of discrete states. For each state, a trajectory/time index will be sampled at which dtrajs
        have an occurrences of this state

    Returns
    -------
    indices : np.ndarray
        The sampled index sequence of shape `(N, 2)`.
        Index array with a number of rows equal to N=len(sequence), with rows consisting of a tuple (i, t),
        where i is the index of the trajectory and t is the time index within the trajectory.

    """
    N = len(sequence)
    res = np.zeros((N, 2), dtype=int)
    for t in range(N):
        s = sequence[t]
        ind = indices[s]
        res[t, :] = indices[s][np.random.randint(len(ind)), :] if len(ind) > 0 else -1

    return res


def indices_by_state(indices, nsample, subset=None, replace=True):
    """Samples trajectory/time indices according to the given sequence of states

    Parameters
    ----------
    indices : List[np.ndarray]
        For each state, all trajectory and time indices where this state occurs.
        Each matrix has a number of rows equal to the number of occurrences of the corresponding state,
        with rows consisting of a tuple (i, t), where i is the index of the trajectory and t is the time index
        within the trajectory.
    nsample : int
        Number of samples per state. If replace = False, the number of returned samples per state could be smaller
        if less than nsample indices are available for a state.
    subset : ndarray((n)), optional, default = None
        array of states to be indexed. By default all states in dtrajs will be used
    replace : boolean, optional
        Whether the sample is with or without replacement

    Returns
    -------
    indices : List[np.ndarray]
        List of the sampled indices by state, each state corresponding to an ndarray of shape (N, 2).
        Each element is an index array with a number of rows equal to N=len(sequence), with rows consisting of a
        tuple (i, t), where i is the index of the trajectory and t is the time index within the trajectory.

    """
    # how many states in total?
    n = len(indices)
    # define set of states to work on
    if subset is None:
        subset = np.arange(n)

    # list of states
    res = np.ndarray(len(subset), dtype=object)
    for i, s in enumerate(subset):
        # how many indices are available?
        m_available = indices[s].shape[0]
        # do we have no indices for this state? Then insert empty array.
        if m_available == 0:
            res[i] = np.zeros((0, 2), dtype=int)
        elif replace:
            I = np.random.choice(m_available, nsample, replace=True)
            res[i] = indices[s][I, :]
        else:
            I = np.random.choice(m_available, min(m_available, nsample), replace=False)
            res[i] = indices[s][I, :]

    return res


def indices_by_distribution(indices: List[np.ndarray], distributions, nsample):
    """Samples trajectory/time indices according to the given probability distributions

    Parameters
    ----------
    indices : list of ndarray( (N_i, 2) )
        For each state, all trajectory and time indices where this state occurs.
        Each matrix has a number of rows equal to the number of occurrences of the corresponding state,
        with rows consisting of a tuple (i, t), where i is the index of the trajectory and t is the time index
        within the trajectory.
    distributions : list or array of ndarray ( (n) )
        m distributions over states. Each distribution must be of length n and must sum up to 1.0
    nsample : int
        Number of samples per distribution. If replace = False, the number of returned samples per state could be smaller
        if less than nsample indices are available for a state.

    Returns
    -------
    indices : length m list of ndarray( (nsample, 2) )
        List of the sampled indices by distribution.
        Each element is an index array with a number of rows equal to nsample, with rows consisting of a
        tuple (i, t), where i is the index of the trajectory and t is the time index within the trajectory.

    """
    # how many states in total?
    n = len(indices)
    for dist in distributions:
        if len(dist) != n:
            raise ValueError('Size error: Distributions must all be of length n (number of states).')

    # list of states
    res = np.ndarray(len(distributions), dtype=object)
    for i, dist in enumerate(distributions):
        # sample states by distribution
        sequence = np.random.choice(n, size=nsample, p=dist)
        res[i] = indices_by_sequence(indices, sequence)
    return res


def by_state(dtrajs, n_samples, subset=None, replace=True):
    """Generates samples of the connected states.

    For each state in the active set of states, generates nsample samples with trajectory/time indices.

    Parameters
    ----------
    dtrajs : List[np.ndarray]
        underlying discrete trajectories
    n_samples : int
        Number of samples per state. If replace = False, the number of returned samples per state could be smaller
        if less than nsample indices are available for a state.
    subset : ndarray((n)), optional, default = None
        array of states to be indexed. By default all states in the connected set will be used
    replace : boolean, optional
        Whether the sample is with or without replacement

    Returns
    -------
    indices : list of ndarray( (N, 2) )
        list of trajectory/time index arrays with an array for each state.
        Within each index array, each row consist of a tuple (i, t), where i is
        the index of the trajectory and t is the time index within the trajectory.
    """
    # generate connected state indices
    indices = compute_index_states(dtrajs, subset=subset)
    return indices_by_state(indices, n_samples, subset=subset, replace=replace)



import warnings
import numpy as np
import scipy

from deeptime.util.exceptions import NotConvergedWarning
from .. import _mle_sparse_bindings as _bindings


def mle_trev(C, maxerr=1.0E-12, maxiter=int(1.0E6),
             warn_not_converged=True, return_statdist=False,
             eps_mu=1.0E-15):
    assert maxerr > 0, 'maxerr must be positive'
    assert maxiter > 0, 'maxiter must be positive'
    assert C.shape[0] == C.shape[1], 'C must be a square matrix.'
    from deeptime.markov.tools.estimation import is_connected
    assert is_connected(C, directed=True), 'C must be strongly connected'

    dtype = C.dtype
    if dtype not in (np.float32, np.float64, np.longdouble):
        dtype = np.float64

    C_sum_py = C.sum(axis=1).A1
    C_sum = C_sum_py.astype(dtype, order='C', copy=False)

    CCt = C + C.T
    # convert CCt to coo format
    CCt_coo = CCt.tocoo()
    n_data = CCt_coo.nnz
    CCt_data = CCt_coo.data.astype(dtype, order='C', copy=False)
    i_indices = CCt_coo.row.astype(int, order='C', copy=True)
    j_indices = CCt_coo.col.astype(int, order='C', copy=True)

    # prepare data array of T in coo format
    T_data = np.zeros(n_data, dtype=dtype, order='C')
    mu = np.zeros(C.shape[0], dtype=dtype, order='C')
    code = _bindings.mle_trev_sparse(T_data, CCt_data, i_indices, j_indices, n_data, C_sum, CCt.shape[0],
                                     maxerr, maxiter, mu, eps_mu)
    if code == -5 and warn_not_converged:
        warnings.warn("Reversible transition matrix estimation with fixed stationary distribution didn't converge.",
                      NotConvergedWarning)

    # T matrix has the same shape and positions of nonzero elements as CCt
    T = scipy.sparse.csr_matrix((T_data, (i_indices, j_indices)), shape=CCt.shape)
    from deeptime.markov.tools.estimation.sparse.transition_matrix import correct_transition_matrix
    T = correct_transition_matrix(T)
    if return_statdist:
        return T, mu
    else:
        return T


def mle_trev_given_pi(C, mu, maxerr=1.0E-12, maxiter=1000000, warn_not_converged=True):
    assert maxerr > 0, 'maxerr must be positive'
    assert maxiter > 0, 'maxiter must be positive'
    from deeptime.markov.tools.estimation import is_connected
    assert is_connected(C, directed=False), 'C must be (weakly) connected'
    dtype = C.dtype
    if dtype not in (np.float32, np.float64, np.longdouble):
        dtype = np.float64
    c_mu = mu.astype(dtype, order='C', copy=False)
    CCt_coo = (C + C.T).tocoo()
    assert CCt_coo.shape[0] == CCt_coo.shape[1] == c_mu.shape[0], 'Dimensions of C and mu don\'t agree.'
    n_data = CCt_coo.nnz
    CCt_data = CCt_coo.data.astype(dtype, order='C', copy=False)
    i_indices = CCt_coo.row.astype(np.uint64, order='C', copy=False)
    j_indices = CCt_coo.col.astype(np.uint64, order='C', copy=False)
    # prepare data array of T in coo format
    T_unnormalized_data = np.zeros(n_data, dtype=dtype, order='C')

    code = _bindings.mle_trev_given_pi_sparse(T_unnormalized_data, CCt_data, i_indices, j_indices, n_data, c_mu,
                                              CCt_coo.shape[0], maxerr, maxiter)

    if code == -5 and warn_not_converged:
        warnings.warn("Reversible transition matrix estimation with fixed stationary distribution didn't converge.",
                      NotConvergedWarning)

    # unnormalized T matrix has the same shape and positions of nonzero elements as the C matrix
    T_unnormalized = scipy.sparse.csr_matrix((T_unnormalized_data, (i_indices.copy(), j_indices.copy())),
                                             shape=CCt_coo.shape)
    # finish T by setting the diagonal elements according to the normalization constraint
    rowsum = T_unnormalized.sum(axis=1).A1
    T_diagonal = scipy.sparse.diags(np.maximum(1.0 - rowsum, 0.0), 0)

    return T_unnormalized + T_diagonal



r"""
.. moduleauthor:: B.Trendelkamp-Schroer <benjamin DOT trendelkamp-schroer AT fu-berlin DOT de>

"""

import numpy as np
from scipy.sparse import issparse, csr_matrix, diags
from scipy.sparse.linalg import minres, LinearOperator


class AugmentedSystem(LinearOperator):

    def __init__(self, J, G, SIG, A):
        self.N1 = J.shape[0]
        self.N2 = A.shape[0]
        N = self.N1 + self.N2
        self.shape = (N, N)
        self.dtype = J.dtype

        self.E = ((G.T.dot(SIG)).dot(G)).tocsr()
        self.J = J
        self.A = A
        self.AT = A.T.tocsr()
        self.diag = np.hstack((J.diagonal() + self.E.diagonal(),
                               np.zeros(self.N2)))

    def _matvec(self, x):
        N1 = self.N1
        N2 = self.N2
        y = np.zeros(N1+N2)
        y[0:N1] = self.J.dot(x[0:N1]) + self.E.dot(x[0:N1]) + self.AT.dot(x[N1:])
        y[N1:] = self.A.dot(x[0:N1])
        return y

    def diagonal(self):
        return self.diag

def mydot(A, B):
    r"""Dot-product that can handle dense and sparse arrays

    Parameters
    ----------
    A : numpy ndarray or scipy sparse matrix
        The first factor
    B : numpy ndarray or scipy sparse matrix
        The second factor

    Returns
    C : numpy ndarray or scipy sparse matrix
        The dot-product of A and B

    """
    if issparse(A) :
        return A.dot(B)
    elif issparse(B):
        return (B.T.dot(A.T)).T
    else:
        return np.dot(A, B)

def factor_aug(z, DPhival, G, A):
    r"""Set up augmented system and return.

    Parameters
    ----------
    z : (N+P+M+M,) ndarray
        Current iterate, z = (x, nu, l, s)
    DPhival : LinearOperator
        Jacobian of the variational inequality mapping
    G : (M, N) ndarray or sparse matrix
        Inequality constraints
    A : (P, N) ndarray or sparse matrix
        Equality constraints

    Returns
    -------
    J : LinearOperator
        Augmented system

    """
    M, N = G.shape
    P, N = A.shape
    """Multiplier for inequality constraints"""
    l = z[N+P:N+P+M]

    """Slacks"""
    s = z[N+P+M:]

    """Sigma matrix"""
    SIG = diags(l/s, 0)
    # SIG = diags(l*s, 0)

    """Convert A"""
    if not issparse(A):
        A = csr_matrix(A)

    """Convert G"""
    if not issparse(G):
        G = csr_matrix(G)

    """Since we expect symmetric DPhival, we need to change A"""
    sign = np.zeros(N)
    sign[0:N//2] = 1.0
    sign[N//2:] = -1.0
    T = diags(sign, 0)

    A_new = A.dot(T)

    W = AugmentedSystem(DPhival, G, SIG, A_new)
    return W

def solve_factorized_aug(z, Fval, LU, G, A):
    M, N=G.shape
    P, N=A.shape

    """Total number of inequality constraints"""
    m=M

    """Primal variable"""
    x=z[0:N]

    """Multiplier for equality constraints"""
    nu=z[N:N+P]

    """Multiplier for inequality constraints"""
    l=z[N+P:N+P+M]

    """Slacks"""
    s=z[N+P+M:]

    """Dual infeasibility"""
    rd = Fval[0:N]

    """Primal infeasibility"""
    rp1 = Fval[N:N+P]
    rp2 = Fval[N+P:N+P+M]

    """Centrality"""
    rc = Fval[N+P+M:]

    """Sigma matrix"""
    SIG = diags(l/s, 0)

    """LU is actually the augmented system W"""
    W = LU

    b1 = -rd - mydot(G.T, mydot(SIG, rp2)) + mydot(G.T, rc/s)
    b2 = -rp1
    b = np.hstack((b1, b2))

    """Prepare iterative solve via MINRES"""
    sign = np.zeros(N+P)
    sign[0:N//2] = 1.0
    sign[N//2:] = -1.0
    T = diags(sign, 0)

    """Change rhs"""
    b_new = mydot(T, b)

    dW = np.abs(W.diagonal())
    dPc = np.ones(W.shape[0])
    ind = (dW > 0.0)
    dPc[ind] = 1.0/dW[ind]
    Pc = diags(dPc, 0)
    dxnu, info = minres(W, b_new, tol=1e-10, M=Pc)

    # dxnu = solve(J, b)
    dx = dxnu[0:N]
    dnu = dxnu[N:]

    """Obtain search directions for l and s"""
    ds = -rp2 - mydot(G, dx)
    # ds = s*ds
    # SIG = np.diag(l/s)
    dl = -mydot(SIG, ds) - rc/s

    dz = np.hstack((dx, dnu, dl, ds))
    return dz



import abc

import numpy as np

from ..util import confidence_interval, LaggedModelValidator
from ..util.types import ensure_array
from ..base import Estimator, Model


class _MSMBaseEstimator(Estimator, metaclass=abc.ABCMeta):
    r"""Maximum likelihood estimator for MSMs given discrete trajectory statistics.
    """

    def __init__(self, reversible=True, sparse=False):
        r""" Creates a new base estimator instance.

        Parameters
        ----------
        reversible : bool, optional, default = True
            If true compute reversible MarkovStateModel, else non-reversible MarkovStateModel
        sparse : bool, optional, default = False
            If true compute count matrix, transition matrix and all derived
            quantities using sparse matrix algebra. In this case python sparse
            matrices will be returned by the corresponding functions instead of
            numpy arrays. This behavior is suggested for very large numbers of
            states (e.g. > 4000) because it is likely to be much more efficient.
        """
        super(_MSMBaseEstimator, self).__init__()
        self.reversible = reversible
        self.sparse = sparse

    @property
    def reversible(self) -> bool:
        r""" If true compute reversible MarkovStateModel, else non-reversible MarkovStateModel """
        return self._reversible

    @reversible.setter
    def reversible(self, value: bool):
        self._reversible = value

    @property
    def sparse(self) -> bool:
        r"""  If true compute count matrix, transition matrix and all derived quantities using sparse matrix algebra.
        In this case python sparse matrices will be returned by the corresponding functions instead of numpy arrays.
        This behavior is suggested for very large numbers of states (e.g. > 4000) because it is likely
        to be much more efficient.
        """
        return self._sparse

    @sparse.setter
    def sparse(self, value: bool):
        self._sparse = value


class BayesianPosterior(Model):
    r""" Bayesian posterior from bayesian MSM sampling.

    Parameters
    ----------
    prior : deeptime.markov.msm.MarkovStateModel, optional, default=None
        The prior.
    samples : list of deeptime.markov.msm.MarkovStateModel, optional, default=None
        Sampled models.

    See Also
    --------
    deeptime.markov.msm.BayesianMSM
    """

    def __init__(self, prior=None, samples=None):
        super().__init__()
        self._prior = prior
        self._samples = samples

    @property
    def samples(self):
        r"""
        The sampled models

        Returns
        -------
        models : list of deeptime.markov.msm.MarkovStateModel or None
            samples
        """
        return self._samples

    @property
    def prior(self):
        r"""
        The prior model.

        Returns
        -------
        prior : deeptime.markov.msm.MarkovStateModel or None
            the prior
        """
        return self._prior

    def __iter__(self):
        for s in self.samples:
            yield s

    def submodel(self, states: np.ndarray):
        r""" Creates a bayesian posterior that is restricted onto the specified states.

        Parameters
        ----------
        states: (N,) ndarray, dtype=int
            array of integers specifying the states to restrict to

        Returns
        -------
        submodel : BayesianPosterior
            A posterior with prior and samples restricted to specified states.
        """
        return BayesianPosterior(
            prior=self.prior.submodel(states),
            samples=[sample.submodel(states) for sample in self.samples]
        )

    def gather_stats(self, quantity, store_samples=False, delimiter='/', confidence=0.95, *args, **kwargs):
        """ Obtain statistics about a sampled quantity. Can also be a chained call, separated by the delimiter.

        Parameters
        ----------
        quantity: str
            name of attribute, which will be evaluated on samples
        store_samples: bool, optional, default=False
            whether to store the samples (array).
        delimiter : str, optional, default='/'
            separator to call members of members
        confidence : float, optional, default=0.95
            Size of the confidence intervals.
        *args
            pass-through
        **kwargs
            pass-through

        Returns
        -------
        statistics : deeptime.util.QuantityStatistics
            The statistics
        """
        from deeptime.util import QuantityStatistics
        return QuantityStatistics.gather(self.samples, quantity=quantity, store_samples=store_samples,
                                         delimiter=delimiter, confidence=confidence, *args, **kwargs)

    def evaluate_samples(self, quantity, delimiter='/', *args, **kwargs):
        r""" Obtains a quantity (like an attribute or result of a method or a property) from each of the samples.
        Returns as list.

        Parameters
        ----------
        quantity : str
            The quantity. Can be also deeper in the instance hierarchy, indicated by the delimiter.
        delimiter : str, default='/'
            The delimiter.
        *args
            Arguments passed to the evaluation point of the quantity.
        **kwargs
            Keyword arguments passed to the evaluation point of the quantity.

        Returns
        -------
        result : list of any or ndarray
            A list of the quantity evaluated on each of the samples. If can be converted to float ndarray then ndarray.
        """
        from deeptime.util.stats import evaluate_samples as _eval
        return _eval(self.samples, quantity=quantity, delimiter=delimiter, *args, **kwargs)


class MembershipsChapmanKolmogorovValidator(LaggedModelValidator):
    r""" Validates a model estimated at lag time tau by testing its predictions for longer lag times.
    This is known as the Chapman-Kolmogorov test as it is based on the Chapman-Kolmogorov equation.

    The test is performed on metastable sets of states rather than the microstates themselves.

    Parameters
    ----------
    memberships : ndarray(n, m)
        Set memberships to calculate set probabilities. n must be equal to
        the number of active states in model. m is the number of sets.
        memberships must be a row-stochastic matrix (the rows must sum up to 1).
    test_model : Model
        Model to be tested
    test_estimator : Estimator
        Parametrized Estimator that has produced the model
    mlags : int or int-array
        Multiples of lag times for testing the Model, e.g. range(10).
        A single int will trigger a range, i.e. `mlags=10` maps to
        `mlags=range(10)`. The setting None will choose mlags automatically
        according to the longest available trajectory.
        Note that you need to be able to do a model prediction for each
        of these lag time multiples, e.g. the value 0 only make sense
        if _predict_observables(0) will work.
    conf : float, default = 0.95
        confidence interval for errors

    Notes
    -----
    This is an adaption of the Chapman-Kolmogorov Test described in detail
    in :footcite:`prinz2011markov` to Hidden MSMs as described in :footcite:`noe2013projected`.

    References
    ----------
    .. footbibliography::
    """

    def __init__(self, test_model, test_estimator, memberships, test_model_lagtime,
                 mlags, conf=0.95):
        super().__init__(test_model, test_estimator, test_model_lagtime=test_model_lagtime, conf=conf, mlags=mlags)
        self.memberships = memberships
        self.test_model = test_model

    @property
    def test_model(self):
        return self._test_model

    @test_model.setter
    def test_model(self, test_model):
        assert self.memberships is not None
        if hasattr(test_model, 'prior'):
            m = test_model.prior
        else:
            m = test_model
        if hasattr(m, 'transition_model'):
            m = m.transition_model
        n_states = m.n_states
        statdist = m.stationary_distribution
        symbols = m.count_model.state_symbols
        symbols_full = m.count_model.n_states_full
        assert self.memberships.shape[0] == n_states, 'provided memberships and test_model n_states mismatch'
        self._test_model = test_model
        # define starting distribution
        P0 = self.memberships * statdist[:, None]
        P0 /= P0.sum(axis=0)  # column-normalize
        self.P0 = P0

        # map from the full set (here defined by the largest state index in active set) to active
        self._full2active = np.zeros(np.max(symbols_full) + 1, dtype=int)
        self._full2active[symbols] = np.arange(len(symbols))

    @property
    def memberships(self):
        return self._memberships

    @memberships.setter
    def memberships(self, value):
        self._memberships = ensure_array(value, ndim=2, dtype=float)
        self.nstates, self.nsets = self._memberships.shape
        assert np.allclose(self._memberships.sum(axis=1), np.ones(self.nstates))  # stochastic matrix?

    def _compute_observables(self, model, mlag):
        if mlag == 0 or model is None:
            return np.eye(self.nsets)
        # otherwise compute or predict them by model.propagate
        pk_on_set = np.zeros((self.nsets, self.nsets))
        # compute observable on prior in case for Bayesian models.
        if hasattr(model, 'prior'):
            model = model.prior
        if hasattr(model, 'transition_model'):
            model = model.transition_model
        symbols = model.count_model.state_symbols
        subset = self._full2active[symbols]  # find subset we are now working on
        for i in range(self.nsets):
            p0 = self.P0[:, i]  # starting distribution on reference active set
            p0sub = p0[subset]  # map distribution to new active set
            if subset is not None:
                p0sub /= p0sub.sum()  # renormalize
            pksub = model.propagate(p0sub, mlag)
            for j in range(self.nsets):
                pk_on_set[i, j] = np.dot(pksub, self.memberships[subset, j])  # map onto set
        return pk_on_set

    def _compute_observables_conf(self, model, mlag, conf=0.95):
        # otherwise compute or predict them by model.propagate
        if mlag == 0 or model is None:
            return np.eye(self.nsets), np.eye(self.nsets)
        prior = model.prior
        if hasattr(prior, 'transition_model'):
            symbols = prior.transition_model.count_model.state_symbols
        else:
            symbols = prior.count_model.state_symbols
        subset = self._full2active[symbols]  # find subset we are now working on
        l = np.zeros((self.nsets, self.nsets))
        r = np.zeros((self.nsets, self.nsets))
        for i in range(self.nsets):
            p0 = self.P0[:, i]  # starting distribution
            p0sub = p0[subset]  # map distribution to new active set
            p0sub /= p0sub.sum()  # renormalize
            pksub_samples = []
            for m in model.samples:
                if hasattr(m, 'transition_model'):
                    m = m.transition_model
                pksub_samples.append(m.propagate(p0sub, mlag))
            for j in range(self.nsets):
                pk_on_set_samples = np.fromiter((np.dot(pksub, self.memberships[subset, j])
                                                 for pksub in pksub_samples), dtype=np.float32,
                                                count=len(pksub_samples))
                l[i, j], r[i, j] = confidence_interval(pk_on_set_samples, conf=self.conf)
        return l, r



"""
Created on May 26, 2014

@author: marscher
"""


class SpectralWarning(RuntimeWarning):
    pass


class ImaginaryEigenValueWarning(SpectralWarning):
    pass


class PrecisionWarning(RuntimeWarning):
    r"""
    This warning indicates that some operation in your code leads
    to a conversion of data-types, which involves a loss/gain in
    precision.
    """
    pass


class NotConvergedWarning(RuntimeWarning):
    r"""
    This warning indicates that some iterative procedure has not
    converged or reached the maximum number of iterations implemented
    as a safe guard to prevent arbitrary many iterations in loops with
    a conditional termination criterion.
    """


class NotConvergedError(RuntimeError):
    pass





from functools import wraps
from typing import Union, Optional, List, Callable

import numpy as np
import scipy
from scipy.sparse import coo_matrix, issparse

from .tools import estimation as msmest
from .tools.analysis import is_connected
from ..base import Model, EstimatorTransformer
from ..util.matrix import submatrix
from ..util.types import ensure_dtraj_list

__author__ = 'noe, clonker'


def requires_state_histogram(func: Callable) -> Callable:
    """
    Decorator marking properties which can only be evaluated if the TransitionCountModel contains statistics over the
    data, i.e., a state histogram.

    Parameters
    ----------
    func : Callable
        the to-be decorated property

    Returns
    -------
    wrapper : Callable
        the decorated property
    """
    @wraps(func)
    def wrap(self, *args, **kw):
        if self.state_histogram is None:
            raise RuntimeError("The model was not provided with a state histogram, this property cannot be evaluated.")
        return func(self, *args, **kw)

    return wrap


class TransitionCountModel(Model):
    r""" Statistics, count matrices, and connectivity from discrete trajectories. These statistics can be used to, e.g.,
    construct MSMs. This model can create submodels (see :func:`submodel`)
    that are restricted to a certain selection of states. This subselection can be made by

    * analyzing the connected sets of the
      count matrix (:func:`connected_sets`)
    * pruning states by thresholding with a mincount_connectivity parameter,
    * or simply providing a subset of states manually.

    Parameters
    ----------
    count_matrix : (N, N) ndarray or sparse matrix
        The count matrix. In case it was estimated with 'sliding', it contains a factor of `lagtime` more counts
        than are statistically uncorrelated.
    counting_mode : str, optional, default=None
        If not None, one of 'sliding', 'sample', or 'effective'.
        Indicates the counting method that was used to estimate the count matrix. In case of 'sliding', a sliding
        window of the size of the lagtime was used to count transitions. It therefore contains a factor
        of `lagtime` more counts than are statistically uncorrelated. It's fine to use this matrix for maximum
        likelihood estimation, but it will give far too small errors if you use it for uncertainty calculations.
        In order to do uncertainty calculations, use the effective count matrix, see
        :attr:`effective_count_matrix`, divide this count matrix by tau, or use 'effective' as estimation parameter.
    lagtime : int, optional, default=1
        The time offset which was used to count transitions in state.
    state_histogram : array_like, optional, default=None
        Histogram over the visited states in discretized trajectories.
    state_symbols : array_like, optional, default=None
        Symbols of the original discrete trajectory that are represented in the counting model. If None, the
        symbols are assumed to represent the data, i.e., a iota range over the number of states. Subselection
        of the model also subselects the symbols.
    count_matrix_full : array_like, optional, default=None
        Count matrix for all state symbols. If None, the count matrix provided as first argument is assumed to
        take that role.
    state_histogram_full : array_like, optional, default=None
        Histogram over all state symbols. If None, the provided state_histogram  is assumed to take that role.

    See Also
    --------
    TransitionCountEstimator
    """

    def __init__(self, count_matrix, counting_mode: Optional[str] = None, lagtime: int = 1,
                 state_histogram: Optional[np.ndarray] = None,
                 state_symbols: Optional[np.ndarray] = None, count_matrix_full=None,
                 state_histogram_full: Optional[np.ndarray] = None):
        super().__init__()
        if count_matrix is None:
            raise ValueError("count matrix was None")
        if not issparse(count_matrix):
            count_matrix = np.asarray(count_matrix)
        if (not issparse(count_matrix) and np.any(count_matrix < 0)) or count_matrix.min() < 0:
            raise ValueError("Count matrix elements must all be non-negative")

        self._count_matrix = count_matrix
        self._counting_mode = counting_mode
        self._lag = lagtime
        self._state_histogram = state_histogram

        if state_symbols is None:
            # if symbols is not set, assume that the count matrix represents all states in the data
            state_symbols = np.arange(self.n_states)

        if len(state_symbols) != self.n_states:
            raise ValueError("Number of symbols in counting model must coincide with the number of states in the "
                             "count matrix! (#symbols = {}, #states = {})".format(len(state_symbols), self.n_states))
        self._state_symbols = state_symbols
        if count_matrix_full is None:
            count_matrix_full = count_matrix
        self._count_matrix_full = count_matrix_full
        if self.n_states_full < self.n_states:
            # full number of states must be at least as large as n_states
            raise ValueError("Number of states was bigger than full number of "
                             "states. (#states = {}, #states_full = {}), likely a wrong "
                             "full count matrix.".format(self.n_states, self.n_states_full))
        if state_histogram_full is None:
            state_histogram_full = state_histogram
        if state_histogram_full is not None and self.n_states_full != len(state_histogram_full):
            raise ValueError(
                "Mismatch between number of states represented in full state histogram and full count matrix "
                "(#states histogram = {}, #states matrix = {})".format(len(state_histogram_full), self.n_states_full)
            )
        self._state_histogram_full = state_histogram_full

    @property
    def state_histogram_full(self) -> Optional[np.ndarray]:
        r""" Histogram over all states in the trajectories. """
        return self._state_histogram_full

    @property
    def n_states_full(self) -> int:
        r""" Full number of states represented in the underlying data. """
        return self.count_matrix_full.shape[0]

    @property
    def states(self) -> np.ndarray:
        r""" The states in this model, i.e., a iota range from 0 (inclusive) to :meth:`n_states` (exclusive).
        See also: :meth:`state_symbols`. """
        return np.arange(self.n_states)

    @property
    def state_symbols(self) -> np.ndarray:
        r""" Symbols (states) that are represented in this count model. """
        return self._state_symbols

    @property
    def counting_mode(self) -> Optional[str]:
        """ The counting mode that was used to estimate the contained count matrix.
        One of 'None', 'sliding', 'sample', 'effective'.
        """
        return self._counting_mode

    @property
    def lagtime(self) -> int:
        """ The lag time at which the Markov model was estimated."""
        return self._lag

    @property
    def is_full_model(self) -> bool:
        r""" Determine whether this counting model refers to the full model that represents all states of the data.

        Returns
        -------
        is_full_model : bool
            Whether this counting model represents all states of the data.
        """
        return self.n_states == self.n_states_full

    def transform_discrete_trajectories_to_submodel(self, dtrajs):
        r"""A list of integer arrays with the discrete trajectories mapped to the currently used set of symbols.
        For example, if there has been a subselection of the model for connectivity='largest', the indices will be
        given within the connected set, frames that do not correspond to a considered symbol are set to -1.

        Parameters
        ----------
        dtrajs : array_like or list of array_like
            discretized trajectories

        Returns
        -------
        array_like or list of array_like
            Curated discretized trajectories so that unconsidered symbols are mapped to -1.
        """

        if self.is_full_model:
            # no-op
            return dtrajs
        else:
            dtrajs = ensure_dtraj_list(dtrajs)
            mapping = -1 * np.ones(self.n_states_full, dtype=np.int32)
            mapping[self.state_symbols] = np.arange(self.n_states)
            return [mapping[dtraj] for dtraj in dtrajs]

    @property
    def count_matrix(self):
        """The count matrix, possibly restricted to a subset of states.

        Attention: This count matrix could have been obtained by sliding a window of length tau across the data.
        It then contains a factor of tau more counts than are statistically uncorrelated. It's fine to use this matrix
        for maximum likelihood estimation, but it will give far too small errors if you use it for uncertainty
        calculations. In order to do uncertainty calculations, use effective counting during estimation,
        or divide this count matrix by tau.
        """
        return self._count_matrix

    @property
    def count_matrix_full(self):
        r""" The count matrix on full set of discrete states, irrespective as to whether they are selected or not.
        """
        return self._count_matrix_full

    @property
    def selected_state_fraction(self) -> float:
        """The fraction of states represented in this count model."""
        return float(self.n_states) / float(self.n_states_full)

    @property
    @requires_state_histogram
    def selected_count_fraction(self) -> float:
        """The fraction of counts represented in this count model."""
        return float(np.sum(self.state_histogram)) / float(np.sum(self.state_histogram_full))

    @property
    def n_states(self) -> int:
        """Number of states """
        return self.count_matrix.shape[0]

    @property
    @requires_state_histogram
    def total_count(self) -> int:
        """Total number of counts"""
        return self._state_histogram.sum()

    @property
    @requires_state_histogram
    def visited_set(self) -> np.ndarray:
        """ The set of visited states. """
        return np.argwhere(self.state_histogram > 0)[:, 0]

    @property
    def state_histogram(self) -> Optional[np.ndarray]:
        """ Histogram of discrete state counts, can be None in case no statistics were provided """
        return self._state_histogram

    def connected_sets(self, connectivity_threshold: float = 0., directed: bool = True,
                       probability_constraint: Optional[np.ndarray] = None,
                       sort_by_population: bool = False) -> List[np.ndarray]:
        r""" Computes the connected sets of the counting matrix. A threshold can be set fixing a number of counts
        required to consider two states connected. In case of sliding window the number of counts is increased by a
        factor of `lagtime`. In case of 'sliding-effective' counting, the number of sliding window counts were
        divided by the lagtime and can therefore also be in the open interval (0, 1). Same for 'effective' counting.

        Parameters
        ----------
        connectivity_threshold : float, optional, default=0.
            Number of counts required to consider two states connected. When the count matrix was estimated with
            effective mode or sliding-effective mode, a threshold of :math:`1 / \mathrm{n_states_full}` is
            commonly used.
        directed : bool, optional, default=True
            Compute connected set for directed or undirected transition graph, default directed
        probability_constraint : (N,) ndarray, optional, default=None
            constraint on the whole state space, sets all counts to zero which have no probability
        sort_by_population : bool, optional, default=False
            This flag can be used to order the resulting list of sets in decreasing order by the most counts.

        Returns
        -------
        A list of arrays containing integers (states), each array representing a connected set. The list is
        ordered decreasingly by the size of the individual components.
        """
        from deeptime.markov import compute_connected_sets
        count_matrix = self.count_matrix
        if probability_constraint is not None:
            # pi has to be defined on all states visited by the trajectories
            if len(probability_constraint) != self.n_states_full:
                raise ValueError("The connected sets with a constraint can only be evaluated if the constraint "
                                 "refers to the whole state space (#states total = {}), but it had a length of "
                                 "#constrained states = {}".format(self.n_states_full, len(probability_constraint)))
            probability_constraint = probability_constraint[self.state_symbols]

            # Find visited states with positive stationary probabilities
            pos = np.where(probability_constraint <= 0.0)[0]
            count_matrix = count_matrix.copy()

            if scipy.sparse.issparse(count_matrix):
                count_matrix = count_matrix.tolil()
            count_matrix[pos, :] = 0.
            count_matrix[:, pos] = 0.
        connected_sets = compute_connected_sets(count_matrix, connectivity_threshold, directed=directed)
        if sort_by_population:
            score = np.array([self.count_matrix[np.ix_(s, s)].sum() for s in connected_sets])
            # want decreasing order therefore sort by -1 * score
            connected_sets = [connected_sets[i] for i in np.argsort(-score)]
        return connected_sets

    def states_to_symbols(self, states: np.ndarray) -> np.ndarray:
        r""" Converts a list of states to a list of symbols which can be related back to the original data.

        Parameters
        ----------
        states : (N,) ndarray
            Array of states.

        Returns
        -------
        symbols : (N,) ndarray
            Array of symbols.
        """
        return self.state_symbols[states]

    def symbols_to_states(self, symbols):
        r"""
        Converts a set of symbols to state indices in this count model instance. The symbols which
        are no longer present in this model are discarded. It can happen that the order is
        changed or the result is smaller than the input length.

        Parameters
        ----------
        symbols : array_like
            the symbols to be mapped to state indices

        Returns
        -------
        states : ndarray
            An array of states.
        """
        # only take symbols which are still present in this model
        if isinstance(symbols, set):
            symbols = list(symbols)
        symbols = np.intersect1d(np.asarray(symbols), self.state_symbols)
        return np.argwhere(np.isin(self.state_symbols, symbols)).flatten()

    def submodel(self, states: np.ndarray):
        r"""This returns a count model that is restricted to a selection of states.

        Parameters
        ----------
        states : array_like
            The states to restrict to.

        Returns
        -------
        submodel : TransitionCountModel
            A submodel restricted to the requested states.
        """
        states = np.atleast_1d(states)
        if np.max(states) >= self.n_states:
            raise ValueError("Tried restricting model to states that are not represented! "
                             "States range from 0 to {}.".format(np.max(states)))
        sub_count_matrix = submatrix(self.count_matrix, states)
        if self.state_symbols is not None:
            sub_symbols = self.state_symbols[states]
        else:
            sub_symbols = None
        if self.state_histogram is not None:
            sub_state_histogram = self.state_histogram[states]
        else:
            sub_state_histogram = None
        return TransitionCountModel(sub_count_matrix, self.counting_mode, self.lagtime, sub_state_histogram,
                                    state_symbols=sub_symbols, count_matrix_full=self.count_matrix_full,
                                    state_histogram_full=self.state_histogram_full)

    def submodel_largest(self, connectivity_threshold: Union[str, float] = 0., directed: Optional[bool] = None,
                         probability_constraint: Optional[np.ndarray] = None, sort_by_population: bool = False):
        r"""
        Restricts this model to the submodel corresponding to the largest connected set of states after eliminating
        states that fall below the specified connectivity threshold.
        
        Parameters
        ----------
        connectivity_threshold : float or '1/n', optional, default=0.
            Connectivity threshold. counts that are below the specified value are disregarded when finding connected
            sets. In case of '1/n', the threshold gets resolved to :math:`1 / n\_states\_full`.
        directed : bool, optional, default=None
            Whether to look for connected sets in a directed graph or in an undirected one. Per default it looks whether
            a probability constraint is given. In case it is given it defaults to the undirected case, otherwise
            directed.
        probability_constraint : (N,) ndarray, optional, default=None
            Constraint on the whole state space (n_states_full). Only considers states that have positive probability.
        sort_by_population : bool, optional, default=False
            This flag can be used to use the connected set with the largest population.

        Returns
        -------
        submodel : TransitionCountModel
            The submodel.
        """
        if directed is None:
            # if probability constraint is given, we want undirected per default
            directed = probability_constraint is None
        if connectivity_threshold == '1/n':
            connectivity_threshold = 1. / self.n_states_full
        connectivity_threshold = float(connectivity_threshold)
        connected_sets = self.connected_sets(connectivity_threshold=connectivity_threshold, directed=directed,
                                             probability_constraint=probability_constraint,
                                             sort_by_population=sort_by_population)
        largest_connected_set = connected_sets[0]
        return self.submodel(largest_connected_set)

    def count_matrix_histogram(self) -> np.ndarray:
        r"""
        Computes a histogram over states represented in the count matrix. The magnitude of the values returned values
        depend on the mode which was used for counting.

        Returns
        -------
        A `(n_states,) np.ndarray` histogram over the collected counts per state.
        """
        return self.count_matrix.sum(axis=1)

    def is_connected(self, directed: bool = True) -> bool:
        r""" Dispatches to :meth:`tools.analysis.is_connected <deeptime.markov.tools.analysis.is_connected>`.
        """
        return is_connected(self.count_matrix, directed=directed)


class TransitionCountEstimator(EstimatorTransformer):
    r"""
    Estimator which produces a :class:`TransitionCountModel` given discretized trajectories.
    Hereby one can decide whether the count mode should be:

        * sample: A trajectory of length T will have :math:`T / \tau` counts at time indices

          .. math:: (0 \rightarrow \tau), (\tau \rightarrow 2 \tau), ..., (((T/ \tau )-1) \tau \rightarrow T)

        * sliding: A trajectory of length T will have :math:`T-\tau` counts at time indices

          .. math:: (0 \rightarrow \tau), (1 \rightarrow \tau+1), ..., (T-\tau-1 \rightarrow T-1)

          This introduces an overestimation of the actual count values by a factor of "lagtime". For
          maximum-likelihood MSMs this plays no role but it leads to wrong error bars in uncertainty estimation.

        * sliding-effective: See sliding mode, just that the resulting count matrix is divided by the lagtime after
          counting. This which can be shown to provide a likelihood that is the geometrical average
          over shifted subsamples of the trajectory, :math:`(s_1,\:s_{tau+1},\:...),\:(s_2,\:t_{tau+2},\:...),` etc.
          This geometrical average converges to the correct likelihood in the statistical limit
          :footcite:`trendelkamp2015estimation`.

        * effective: Uses an estimate of the transition counts that are statistically uncorrelated. Recommended
          when used with a Bayesian MSM. A description of the estimation procedure
          can be found in :footcite:`noe2015statistical`.

    Parameters
    ----------
    lagtime : int
        Distance between two frames in the discretized trajectories under which their potential change of state
        is considered a transition.
    count_mode : str
        One of "sample", "sliding", "sliding-effective", and "effective".

        * "sample" strides the trajectory with lagtime :math:`\tau` and uses the strided counts as transitions.
        * "sliding" uses a sliding window approach, yielding counts that are statistically correlated and too
          large by a factor of :math:`\tau`; in uncertainty estimation this yields wrong uncertainties.
        * "sliding-effective" takes "sliding" and divides it by :math:`\tau`, which can be shown to provide a
          likelihood that is the geometrical average over shifted subsamples of the trajectory,
          :math:`(s_1,\:s_{tau+1},\:...),\:(s_2,\:t_{tau+2},\:...),` etc. This geometrical average converges to
          the correct likelihood in the statistical limit :footcite:`trendelkamp2015estimation`.
        * "effective" uses an estimate of the transition counts that are statistically uncorrelated.
          Recommended when estimating Bayesian MSMs.
    n_states : int, optional, default=None
        Normally, the shape of the count matrix is a consequence of the number of encountered states in given
        discrete trajectories. However sometimes (for instance when scoring), only a portion of the discrete
        trajectories is passed but the count matrix should still have the correct shape. Then, this argument
        can be used to artificially set the number of states to the correct value.
    sparse : bool, optional, default=False
        Whether sparse matrices should be used for counting. This can make sense when the number of states is very
        large.

    See Also
    --------
    TransitionCountModel

    References
    ----------
    .. footbibliography::
    """

    def __init__(self, lagtime: int, count_mode: str, n_states=None, sparse=False):
        super().__init__()
        self.lagtime = lagtime
        self.count_mode = count_mode
        self.sparse = sparse
        self.n_states = n_states

    @property
    def count_mode(self):
        r""" The currently selected count mode. """
        return self._count_mode

    @count_mode.setter
    def count_mode(self, value):
        valids = ('sliding', 'sliding-effective', 'sample', 'effective')
        if value not in valids:
            raise ValueError("Invalid count mode \"{}\", possible values are {}.".format(value, valids))
        self._count_mode = value

    @property
    def sparse(self) -> bool:
        r""" Whether the resulting count matrix is stored in sparse or dense mode.

        :getter: Yields the currently configured sparsity setting.
        :setter: Sets whether to store count matrices sparsely.
        :type: bool
        """
        return self._sparse

    @sparse.setter
    def sparse(self, value: bool):
        self._sparse = bool(value)

    @property
    def n_states(self) -> Optional[bool]:
        r""" The number of states in discrete trajectories. Can be used to override the effective shape
        of the resulting count matrix.

        :getter: Yields the currently set number of states or None.
        :setter: Sets the number of states to use or None.
        :type: bool or None
        """
        return self._n_states

    @n_states.setter
    def n_states(self, value):
        self._n_states = value

    @property
    def lagtime(self) -> int:
        r""" The lagtime at which transitions are counted. """
        return self._lagtime

    @lagtime.setter
    def lagtime(self, value: int):
        self._lagtime = value

    def fetch_model(self) -> Optional[TransitionCountModel]:
        r"""
        Yields the latest estimated :class:`TransitionCountModel`. Might be None if fetched before any data was fit.

        Returns
        -------
        The latest :class:`TransitionCountModel` or None.
        """
        return self._model

    def fit(self, data, *args, **kw):
        r""" Counts transitions at given lag time according to configuration of the estimator.

        Parameters
        ----------
        data : array_like or list of array_like
            discretized trajectories
        """
        from deeptime.markov import count_states
        dtrajs = ensure_dtraj_list(data)

        # basic count statistics
        histogram = count_states(dtrajs, ignore_negative=True)

        # Compute count matrix
        count_mode = self.count_mode
        lagtime = self.lagtime
        count_matrix = TransitionCountEstimator.count(count_mode, dtrajs, lagtime, sparse=self.sparse,
                                                      n_jobs=kw.pop('n_jobs', None))
        if self.n_states is not None and self.n_states > count_matrix.shape[0]:
            histogram = np.pad(histogram, pad_width=[(0, self.n_states - count_matrix.shape[0])])
            if issparse(count_matrix):
                count_matrix = scipy.sparse.csr_matrix((count_matrix.data, count_matrix.indices, count_matrix.indptr),
                                                       shape=(self.n_states, self.n_states))
            else:
                n_pad = self.n_states - count_matrix.shape[0]
                count_matrix = np.pad(count_matrix, pad_width=[(0, n_pad), (0, n_pad)])

        # initially state symbols, full count matrix, and full histogram can be left None because they coincide
        # with the input arguments
        self._model = TransitionCountModel(
            count_matrix=count_matrix, counting_mode=count_mode, lagtime=lagtime, state_histogram=histogram
        )
        return self

    @staticmethod
    def count(count_mode: str, dtrajs: List[np.ndarray], lagtime: int, sparse: bool = False, n_jobs=None):
        r""" Computes a count matrix based on a counting mode, some discrete trajectories, a lagtime, and
        whether to use sparse matrices.

        Parameters
        ----------
        count_mode : str
            The counting mode to use. One of "sample", "sliding", "sliding-effective", and "effective".
            See :meth:`__init__` for a more detailed description.
        dtrajs : array_like or list of array_like
            Discrete trajectories, i.e., a list of arrays which contain non-negative integer values. A single ndarray
            can also be passed, which is then treated as if it was a list with that one ndarray in it.
        lagtime : int
            Distance between two frames in the discretized trajectories under which their potential change of state
            is considered a transition.
        sparse : bool, default=False
            Whether to use sparse matrices or dense matrices. Sparse matrices can make sense when dealing with a lot of
            states.
        n_jobs : int, optional, default=None
            This only has an effect in effective counting. Determines the number of cores to use for estimating
            statistical inefficiencies. Default resolves to number of available cores.

        Returns
        -------
        count_matrix : (N, N) ndarray or sparse array
            The computed count matrix. Can be ndarray or sparse depending on whether sparse was set to true or false.
            N is the number of encountered states, i.e., :code:`np.max(dtrajs)+1`.

        Example
        -------
        >>> dtrajs = [np.array([0,0,1,1]), np.array([0,0,1])]
        >>> count_matrix = TransitionCountEstimator.count(
        ...     count_mode="sliding", dtrajs=dtrajs, lagtime=1, sparse=False
        ... )
        >>> np.testing.assert_equal(count_matrix, np.array([[2, 2], [0, 1]]))
        """
        if count_mode == 'sliding' or count_mode == 'sliding-effective':
            count_matrix = msmest.count_matrix(dtrajs, lagtime, sliding=True, sparse_return=sparse)
            if count_mode == 'sliding-effective':
                count_matrix /= lagtime
        elif count_mode == 'sample':
            count_matrix = msmest.count_matrix(dtrajs, lagtime, sliding=False, sparse_return=sparse)
        elif count_mode == 'effective':
            count_matrix = msmest.effective_count_matrix(dtrajs, lagtime, n_jobs=n_jobs)
            if not sparse and issparse(count_matrix):
                count_matrix = count_matrix.toarray()
        else:
            raise ValueError('Count mode {} is unknown.'.format(count_mode))
        return count_matrix
